import numpy as np
from sklearn.model_selection import GridSearchCV
from sklearn.multioutput import MultiOutputRegressor
import pypce
from tqdm import tqdm
from sklearn.base import BaseEstimator, RegressorMixin, TransformerMixin

from sklearn.utils.estimator_checks import check_estimator
from sklearn.utils.validation import check_X_y, check_array
from sklearn.model_selection import KFold, cross_val_score, cross_validate

from sklearn import linear_model
import warnings
from sklearn.decomposition import PCA
from sklearn.preprocessing import FunctionTransformer

import pandas as pd
import ipdb

n, d = 30, 2
rn = np.random.RandomState(45)
X = 2 * rn.rand(n, d) - 1
y = 3 + 2 * .5 * (3 * X[:, 0] ** 2 - 1) + X[:, 1] * .5 * (3 * X[:, 0] ** 2 - 1)
y2 = 1 + 2 * .5 * (3 * X[:, 1] ** 2 - 1) + 3 * X[:, 0] * .5 * (3 * X[:, 1] ** 2 - 1)
Y = np.vstack([y, y2]).T  # test for multioutput regressor
c_true = np.array([3., 0., 0., 2., 0., 0., 0., 1., 0., 0., ])


class MultiOutputCV(BaseEstimator, RegressorMixin):
    def __init__(self, regressor_model, regressor_param_grid,
                 score_metric='neg_root_mean_squared_error',
                 n_jobs=4,cv=None):
        self.regressor_model = regressor_model
        self.regressor_param_grid = regressor_param_grid
        self.score_metric = score_metric
        self.n_jobs_ = n_jobs
        self.cv = cv
    def _setupCV(self, shuffle=False, randstate=13):
        if self.cv == None:
            self.cv = KFold(n_splits=5)  # ,shuffle=True,random_state=13)
    def fit(self, X, Y):
        self.n, self.dim = X.shape
        if Y.ndim == 1: Y = np.atleast_2d(Y).T
        assert len(Y) == self.n, "mistmatch in no of samples btwn X and Y."
        self.ntargets = Y.shape[1]
        self._setupCV()
        pceCV = GridSearchCV(self.regressor_model, self.regressor_param_grid, scoring=self.score_metric,
            n_jobs=self.n_jobs_, cv=self.cv, 
            verbose=1, return_train_score=True)
        self.best_estimators_ = []
        self.best_params_ = []
        # t = tqdm(range(self.ntargets), desc='target #', leave=True)
        # for i in t:
        for i in range(self.ntargets):
            # t.set_description("target #%i" % (i + 1))
            pceCV.fit(X, Y[:, i])
            self.best_estimators_.append(pceCV.best_estimator_)
            self.best_params_.append(pceCV.best_params_)
        self.pceCV = pceCV
        return self

    def predict(self, X):
        y = []
        for estimator_ in self.best_estimators_:
            y.append(estimator_.predict(X))
        if self.ntargets == 1:
            y = y[0]
        else:
            y = np.vstack(y).T
        return y

    def feature_importances(self):
        S = []
        for estimator_ in self.best_estimators_:
            assert hasattr(estimator_,'feature_importances_')
            S.append(estimator_.sensitivity_indices())
        if self.ntargets == 1:
            S = np.array(S[0])
        else:
            S = np.vstack(S)
        return S

class PCATargetTransform(BaseEstimator, TransformerMixin):
    def __init__(self, n_components=2, cutoff=1e-2, svd_solver='arpack'):
        self.n_components = n_components
        self.cutoff = cutoff
        self.svd_solver = svd_solver
        self.K = None

    def fit(self, Y):
        self.n, self.d = Y.shape
        if self.n_components == 'auto':
            self._compute_K(Y)
        if isinstance(self.n_components, int):
            self.K = self.n_components
            if self.n_components > self.d:
                warnings.warn("No of components greater than dimension. Setting to maximum allowable.")
                self.n_components = self.d
        assert self.K is not None, "K components is not defined or being set properly."
        self.pca = PCA(n_components=self.K, svd_solver=self.svd_solver)
        self.pca.fit(Y)
        return self

    def fit_transform(self, Y):
        self.fit(Y)
        return self.pca.transform(Y)

    def transform(self, Y):
        assert hasattr(self, 'pca'), "Perform fit first."
        return self.pca.transform(Y)

    def inverse_transform(self, Yhat):
        return self.pca.inverse_transform(Yhat)

    def _compute_K(self, Y, max_n_components=50):
        ''' automatically compute number of PCA terms '''
        self.pca = PCA(n_components=min(max_n_components, self.d), svd_solver=self.svd_solver)
        self.pca.fit(Y)
        self.cumulative_error = np.cumsum(self.pca.explained_variance_ratio_)
        self.K = np.where(1 - self.cumulative_error <= self.cutoff)[0][0]

def plot_feature_importance(S,feature_labels,extra_line_plot=None):
    assert isinstance(S,np.ndarray), "S must be a numpy array"
    if S.ndim == 1:
        ntargets = 1
        ndim = len(S)
        S = np.atleast_2d(S)

    ntargets, ndim = S.shape
    # normalize across columns (if not already)
    S = S / np.atleast_2d(S.sum(axis=1)).T

    # plot sobol indices as stacked bar charts

    import matplotlib.pyplot as plt
    import matplotlib._color_data as mcd
    import seaborn as sns

    xkcd_colors = []
    xkcd = {name for name in mcd.CSS4_COLORS if "xkcd:" + name in mcd.XKCD_COLORS}
    for j, n in enumerate(xkcd):
        xkcd = mcd.XKCD_COLORS["xkcd:" + n].upper()
        xkcd_colors.append(xkcd)

    Ps = []
    sns.set_palette(sns.color_palette("Paired", 12))
    plt.figure(figsize=(20, 9))
    bottom = np.zeros(ntargets)
    for ii in range(ndim):
        ptemp = plt.bar(np.arange(1, 1 + ntargets), S[:, ii], bottom=bottom, width=.25)
        bottom = bottom + S[:, ii]  # reset bottom to new height
        Ps.append(ptemp)
    plt.ylabel('Sobol Scores')
    plt.ylim([0, 1.1])
    # plt.title('Sobol indices by pca mode')
    # plt.xticks(t, ('PCA1','PCA2','PCA3','PCA4','PCA5','PCA6'))
    # X_col_names = ['dens_sc', 'vel_sc', 'temp_sc',
    #            'sigk1', 'sigk2', 'sigw1', 'sigw2',
    #            'beta_s', 'kap', 'a1', 'beta1r', 'beta2r']
    plt.legend(((p[0] for p in Ps)), (l for l in feature_labels),
           fancybox=True, shadow=True,
           loc='upper center', bbox_to_anchor=(0.5, 1.05), ncol=ndim)

    # plot explained variance
    if extra_line_plot is not None:
        assert len(extra_line_plot) >= ntargets, "extra info must be the same length as the targets."
        plt.plot(np.arange(1, 1 + ntargets), extra_line_plot[:ntargets], '--ok')
    return plt

print("Performing grid search for multiple target fits one at a time...")
pce_param_grid = [
    {'order': [1, 2, 3],
     'mindex_type': ['total_order', 'hyperbolic', ],
     'fit_type': ['LassoCV', 'linear', 'ElasticNetCV']}
]
model = pypce.pcereg()

pceMCV = MultiOutputCV(regressor_model=model, regressor_param_grid=pce_param_grid,n_jobs=8)


# # test fit with one target
# pceMCV.fit(X,y)

# # test with multiple targets
# pceMCV.fit(X,Y)

# ipdb.set_trace()

################################################################
# PCA Target Transform
################################################################
print("Performing PCA transform...")
# test target transform
# Load the data
datadir = '/Users/kchowdh/Research/sparc_tests/'
X = np.load(datadir + 'data/X_' + 'heat_flux_wall' + '.npy')
Y = np.load(datadir + 'data/Y_' + 'heat_flux_wall' + '.npy')

# Scale target Y (use inverse transform for new data)
# Each row is a functional output, so we should scale each row
target_scaler = pypce.MinMaxTargetScaler(target_range=(0, 1))
Y_scaled = target_scaler.fit_transform(Y.T).T

# Lower and Upper Bound for features in X (for future scaling)
X_LB = np.array([.85, .85, .85, 0.7, 0.8, 0.3, 0.7, 0.0784, 0.38, 0.31, 1.19, 1.05])
X_UB = np.array([1.15, 1.15, 1.15, 1.0, 1.2, 0.7, 1.0, 0.1024, 0.42, 0.40, 1.31, 1.45])

X_col_names = ['dens_sc', 'vel_sc', 'temp_sc',
               'sigk1', 'sigk2', 'sigw1', 'sigw2',
               'beta_s', 'kap', 'a1', 'beta1r', 'beta2r']

# Scale features X (use inverse transform for new data)
feature_scaler = pypce.DomainScaler(a=X_LB, b=X_UB, domain_range=(-1, 1))
X_scaled = feature_scaler.fit_transform(X)

pca = PCATargetTransform(n_components='auto', cutoff=1e-3)
pca.fit(Y_scaled)
Yhat_scaled = pca.transform(Y_scaled)

from sklearn.metrics import make_scorer
from collections import defaultdict
cv_test_scores = defaultdict(list)
cv_train_scores = defaultdict(list)
def custom_metric(y_true,y_pred,**kwargs):
    return np.linalg.norm(y_true - y_pred)/np.linalg.norm(y_true)
scorer = make_scorer(custom_metric,greater_is_better=False) 
# scorer = 'neg_root_mean_squared_error'
nprocs = 8
S_pce, S_rf, S_ab, S_gb = [], [], [], []
S_mlp, S_knn = [], []

import time as T
start = T.time()
for nn in range(Yhat_scaled.shape[1]-7):
    print("\nWorking on mode %i ...\n" %(nn+1))
    y = Yhat_scaled[:,nn]
    ################################################################
    # PCE Fit
    ################################################################
    pce_param_grid = [{'order': [1, 2, 3],
                       'mindex_type': ['total_order', 'hyperbolic', ],
                       'fit_type': ['LassoCV', 'linear', 'ElasticNetCV']}
                     ]
    model = pypce.pcereg()

    pceMCV = MultiOutputCV(regressor_model=model, regressor_param_grid=pce_param_grid,n_jobs=nprocs,score_metric=scorer)


    print("\n" + "*"*80 + "\nFitting with PCEs...\n" + "*"*80 + "\n")
    pceMCV.fit(X_scaled, y)
    S = pceMCV.feature_importance()
    S_pce.append(S)
    # plt = plot_feature_importance(S,X_col_names,pca.cumulative_error)
    print("Best PCE fit score: ", pceMCV.pceCV.best_score_)
    cv_test_scores['pce'].append(pceMCV.pceCV.best_score_)
    cv_score_argmin = pceMCV.pceCV.best_index_
    cv_train_scores['pce'].append(pceMCV.pceCV.cv_results_['mean_train_score'][cv_score_argmin])

    ################################################################
    # random forest regressor test
    ################################################################ 
    print("\n" + "*"*80 + "\nFitting with Random Forests...\n" + "*"*80 + "\n")
    from sklearn.ensemble import RandomForestRegressor

    # get cv from above
    cv = pceMCV.pceCV.cv

    # Create the random grid
    # 'min_samples_leaf': [2,.01,.05]
    # 'max_depth': [2,4,8,16] 
    # 'max_samples': [.5,.99],
    rf_param_grid = {'n_estimators': [10000,20000],
                     'max_features': ['log2'],
                     'max_depth': [4,5,6]
                     }
    # scorer = 'r2'
    rfreg = RandomForestRegressor(warm_start=False)
    rfregCV = GridSearchCV(rfreg,rf_param_grid,scoring=scorer,cv=cv,verbose=1,n_jobs=nprocs,return_train_score=True)

    rfregCV.fit(X_scaled,y)
    best_rf_estimator = rfregCV.best_estimator_

    S_rf.append(best_rf_estimator.feature_importances_)
    print("Best RF fit score: ", rfregCV.best_score_)
    cv_test_scores['rf'].append(rfregCV.best_score_)
    cv_score_argmin = rfregCV.best_index_
    cv_train_scores['rf'].append(rfregCV.cv_results_['mean_train_score'][cv_score_argmin])
    depths = [estimator.tree_.max_depth for estimator in best_rf_estimator.estimators_]

    ################################################################
    # adaboost
    ################################################################ 
    print("\n" + "*"*80 + "\nFitting with AdaBoost...\n" + "*"*80 + "\n")
    from sklearn.ensemble import AdaBoostRegressor

    # get cv from above
    cv = pceMCV.pceCV.cv
    scorer_ab = scorer
    # scorer = 'r2'

    ab_param_grid = {'n_estimators': [100,500],
                     'loss': ['linear','square']}
    ABreg = AdaBoostRegressor()
    ABregCV = GridSearchCV(ABreg,ab_param_grid,scoring=scorer,cv=cv,verbose=1,n_jobs=nprocs,return_train_score=True)

    ABregCV.fit(X_scaled,y)
    best_ab_estimator = ABregCV.best_estimator_

    S_ab.append(best_ab_estimator.feature_importances_)
    print("Best AdaBoost fit score: ", ABregCV.best_score_)
    cv_test_scores['ab'].append(ABregCV.best_score_)
    cv_score_argmin = ABregCV.best_index_
    cv_train_scores['ab'].append(ABregCV.cv_results_['mean_train_score'][cv_score_argmin])

    ################################################################
    # gradient boosting regressor
    ################################################################ 
    print("\n" + "*"*80 + "\nFitting with GradientBoost...\n" + "*"*80 + "\n")
    from sklearn.ensemble import GradientBoostingRegressor

    # get cv from above
    cv = pceMCV.pceCV.cv

    gb_param_grid = {'n_estimators': [500],
                     'learning_rate': [.1,.5],
                     'max_depth': [1,2],
                     'loss': ['ls','huber']
                     }
    GBreg = GradientBoostingRegressor()
    GBregCV = GridSearchCV(GBreg,gb_param_grid,scoring=scorer,cv=cv,verbose=1,n_jobs=nprocs,return_train_score=True)

    # X_transformed = pceMCV.pceCV.best_estimator_.fit_transform(X_scaled)

    GBregCV.fit(X_scaled,y)
    best_gb_estimator = GBregCV.best_estimator_

    S_gb.append(best_gb_estimator.feature_importances_)
    print("Best GradBoost fit score: ", GBregCV.best_score_)
    cv_test_scores['gb'].append(GBregCV.best_score_)
    cv_score_argmin = GBregCV.best_index_
    cv_train_scores['gb'].append(GBregCV.cv_results_['mean_train_score'][cv_score_argmin])

    ################################################################
    # MLP regressor
    ################################################################ 
    print("\n" + "*"*80 + "\nFitting with MLP...\n" + "*"*80 + "\n")
    from sklearn.neural_network import MLPRegressor

    # get cv from above
    cv = pceMCV.pceCV.cv

    mlp_param_grid = {'hidden_layer_sizes': [(1000)],
                     'solver': ['lbfgs','adam'],
                     'activation': ['relu'],
                     'max_iter': [1000],
                     'batch_size': ['auto'],
                     'learning_rate': ['constant','adaptive'],
                     'alpha': [.0001],
                     'tol': [1e-5]
                     }
    MLPreg = MLPRegressor()
    MLPregCV = GridSearchCV(MLPreg,mlp_param_grid,scoring=scorer,cv=cv,verbose=1,n_jobs=nprocs,return_train_score=True)

    X_transformed = pceMCV.pceCV.best_estimator_.fit_transform(X_scaled)
    # MLPregCV.fit(X_transformed,y) # exclude bias
    MLPregCV.fit(X_scaled,y)
    best_mlp_estimator = MLPregCV.best_estimator_

    # S_mlp.append(best_mlp_estimator.feature_importances_)
    print("Best MLP fit score: ", MLPregCV.best_score_)
    cv_test_scores['mlp'].append(MLPregCV.best_score_)
    cv_score_argmin = MLPregCV.best_index_
    cv_train_scores['mlp'].append(MLPregCV.cv_results_['mean_train_score'][cv_score_argmin])

    ################################################################
    # Nearest neighbor regressor
    ################################################################ 
    print("\n" + "*"*80 + "\nFitting with MLP...\n" + "*"*80 + "\n")
    from sklearn.neighbors import KNeighborsRegressor

    # get cv from above
    cv = pceMCV.pceCV.cv

    knn_param_grid = {'n_neighbors': [2,5,8,10],
                      'weights': ['uniform','distance'],
                      'algorithm': ['auto','brute'],
                      'leaf_size': [1,2,3,4],
                      'p': [2]
                     }
    kNNreg = KNeighborsRegressor()
    kNNregCV = GridSearchCV(kNNreg,knn_param_grid,scoring=scorer,cv=cv,verbose=1,n_jobs=nprocs,return_train_score=True)

    # X_transformed = pceMCV.pceCV.best_estimator_.fit_transform(X_scaled)
    # kNNregCV.fit(X_transformed,y) # exclude bias
    kNNregCV.fit(X_scaled,y)
    best_knn_estimator = kNNregCV.best_estimator_

    # S_knn.append(best_knn_estimator.feature_importances_)
    print("Best kNN fit score: ", kNNregCV.best_score_)
    cv_test_scores['knn'].append(kNNregCV.best_score_)
    cv_score_argmin = kNNregCV.best_index_
    cv_train_scores['knn'].append(kNNregCV.cv_results_['mean_train_score'][cv_score_argmin])
print("Total time is %.5f" %(T.time() - start))
################################################################
# plot validation scores
################################################################
import plotly.graph_objects as go

def plot_multiple_scatter(x,Y,Y_col_labels=None,ms=20,alpha=1,scale=1.0):
    # x can be a list of strings or a 1d array
    # Y can be a dictionary or a ndarray
    if isinstance(Y,np.ndarray):
        assert len(x) ==  Y.shape[1]
        assert len(Y_col_labels) == Y.shape[0]
        n_scatter = Y.shape[0]
    if isinstance(Y,dict):
        if Y_col_labels is None:
            Y_col_labels = list(Y.keys())
        n_scatter = len(Y_col_labels)
        Y = np.array([Y[k] for k in Y.keys()])
    fig = go.Figure()
    for i,y in enumerate(Y):
        fig.add_trace(go.Scatter(x=x, y=scale*y,
                                mode='lines+markers',
                                marker=dict(size=ms,opacity=alpha),
                                line=dict(width=2,dash='dash'),
                                name=Y_col_labels[i]))
    fig.update_layout(font=dict(size=18))
    return fig


x_labels = ["PCA mode %i" %(n+1) for n in range(len(cv_test_scores['pce']))]
col_labels_custom = ['PCE', 'RandForests', 'AdaBoost', 'GradBoost', 'MLP', 'kNN']
fig1 = plot_multiple_scatter(x_labels,cv_test_scores,Y_col_labels=col_labels_custom,alpha=.7,scale=-1.0)  
fig1.write_image("cv_test_scores.png",width=2000,height=1000)

cv_diff = np.array([np.array(cv_test_scores[k])-np.array(cv_train_scores[k]) for k in cv_train_scores.keys()])
col_labels = [k for k in cv_train_scores.keys()]
fig2 = plot_multiple_scatter(x_labels,cv_diff,Y_col_labels=col_labels_custom,alpha=.7,scale=-1.0)  
fig2.write_image("cv_diff_scores.png",width=2000,height=1000)

################################################################
# use plotly to plot interactive plots
################################################################
# print("Graphing...")
# # Need plotly, psutil, orca*, and pandas
# # * orca is a pain to install (easiest is to go to github and follow dmg installation guide)
# import plotly.express as px
# import plotly.graph_objects as go
# from plotly.subplots import make_subplots
# import pandas as pd
# import numpy as np


# class HiDFuncVisualizer():
#     def __init__(self, X, Y, feature_labels=None, target_labels=None, thin=None):
#         self.X = X
#         self.Y = Y
#         self.feature_labels = feature_labels
#         self.target_labels = target_labels
#         self.thin = thin

#     def _precompile(self):
#         # get dimensiona and shapes
#         self.nx, self.d = X.shape
#         if self.Y.ndim == 1:
#             self.Y = np.atleast_2d(self.Y).T
#         self.ny, self.ntargets = self.Y.shape
#         assert self.nx == self.ny, "X and Y have to have the same number of samples."
#         # check column names
#         if self.feature_labels is None:
#             self.feature_labels = ['x_s' + str(i + 1) for i in range(self.d)]
#         else:
#             assert len(self.feature_labels) == self.d, "feature labels need to be size " + self.d
#         if self.target_labels is None:
#             self.target_labels = ['y_' + str(i + 1) for i in range(self.ntargets)]
#         else:
#             assert len(self.target_labels) == self.ntargets, "target labels are not the right size. "
#         # size of plots
#         self.nrows = int(round(np.sqrt(self.d)))
#         self.ncols = round(self.d / self.nrows)

#     def compile(self):
#         # convert data to pandas dataframe
#         self._precompile()
#         if self.thin == None:
#             self.X_df = pd.DataFrame(data=self.X, columns=self.feature_labels)
#             self.Y_df = pd.DataFrame(data=self.Y, columns=self.target_labels)
#         elif isinstance(self.thin, int):
#             self.X_df = pd.DataFrame(data=self.X[::self.thin, :], columns=self.feature_labels)
#             self.Y_df = pd.DataFrame(data=self.Y[::self.thin, :], columns=self.target_labels)

#     def render(self, target_number=1, nrows=None, ncols=None, title=None):
#         # plot univariate, select which target_number you want to plot
#         if nrows is None and ncols is None:
#             nrows = self.nrows
#             ncols = self.ncols
#         assert nrows * ncols >= self.ntargets, "rows and cols must be bigger"
#         fig = make_subplots(rows=nrows, cols=ncols)
#         ycol = self.target_labels[target_number - 1]
#         for r in range(nrows):
#             for c in range(ncols):
#                 if r * ncols + c >= self.d:
#                     break
#                 else:
#                     xcol = self.feature_labels[r * ncols + c]
#                     fig.add_trace(
#                         go.Scatter(x=self.X_df[xcol],
#                                    y=self.Y_df[ycol],
#                                    mode='markers',
#                                    name=xcol + ' vs. ' + ycol),
#                         row=r + 1, col=c + 1
#                     )
#                 fig.update_xaxes(title_text=xcol, row=r + 1, col=c + 1)
#         fig.update_layout(title_text=ycol)
#         return fig

#     def render_corner(self, target_number=1, n_features=None):
#         # n_features = min(n_features,6)
#         if n_features is None:
#             self.mtx_nrows = len(self.feature_labels)
#             self.mtx_ncols = self.mtx_nrows
#         else:
#             self.mtx_nrows = len(self.feature_labels[:n_features])
#             self.mtx_ncols = self.mtx_nrows
#         if self.mtx_nrows > 6: warnings.warn("n_features > 6 does not render when using fig.show()")
#         # subplots
#         nrows, ncols = self.mtx_nrows, self.mtx_ncols
#         # spec = {'is_3d':True}
#         spec = {'type': 'Surface'}
#         specs = [[spec for j in range(ncols)] for i in range(nrows)]
#         # make diagonal univariate
#         for i in range(nrows):
#             specs[i][i] = {'type': 'Scatter'}
#         # fig = make_subplots(rows=nrows, cols=ncols,specs=specs)

#         fig = make_subplots(rows=nrows, cols=ncols, specs=specs, vertical_spacing=0.01, horizontal_spacing=.01,
#                             row_titles=self.feature_labels[:n], column_titles=self.feature_labels[:n_features])

#         zcol = self.target_labels[target_number - 1]
#         for r in range(1, nrows):
#             count = 0
#             for c in range(min(r, ncols)):
#                 # ix, iy = pairs[count]
#                 xcol1 = self.feature_labels[:n_features][r]
#                 xcol2 = self.feature_labels[:n_features][c]
#                 fig.add_trace(
#                     go.Scatter3d(x=self.X_df[xcol1],
#                                  y=self.X_df[xcol2],
#                                  z=self.Y_df[zcol],
#                                  name=xcol1 + ', ' + xcol2 + ' vs. ' + zcol,
#                                  mode='markers',
#                                  marker=dict(size=4),
#                                  opacity=.65),
#                     row=r + 1, col=c + 1
#                 )
#         for r in range(nrows):
#             c = r
#             xcol1 = self.feature_labels[:n_features][r]
#             xcol2 = self.feature_labels[:n_features][c]
#             fig.add_trace(
#                 go.Scatter(x=self.X_df[xcol1],
#                            y=self.Y_df[zcol],
#                            name=xcol1 + ' vs. ' + zcol,
#                            mode='markers',
#                            opacity=.65),
#                 row=r + 1, col=c + 1
#             )
#         fig.update_layout(title_text=zcol)
#         return fig


# # visualization
# # create a pandas data frame
# X_col_names = ['dens_sc', 'vel_sc', 'temp_sc',
#                'sigk1', 'sigk2', 'sigw1', 'sigw2',
#                'beta_s', 'kap', 'a1', 'beta1r', 'beta2r']
# Yhat_col_names = ['PCA mode #' + str(i) + ' for heat flux' for i in range(1, 1 + Yhat_scaled.shape[1])]

# # convert data to pandas data frame
# X_scaled_df = pd.DataFrame(data=X_scaled, columns=X_col_names)
# Yhat_scaled_df = pd.DataFrame(data=Yhat_scaled, columns=Yhat_col_names)

# # use plotly to plot interactive plots
# import plotly.express as px
# import plotly.graph_objects as go
# from plotly.subplots import make_subplots

# F = HiDFuncVisualizer(X_scaled, Yhat_scaled, feature_labels=X_col_names, target_labels=Yhat_col_names, thin=3)
# F.compile()
# print('saving figures...')
# fig1 = F.render(target_number=1)
# fig1.write_image("univariate.png",width=1200,height=1000)
# fig2 = F.render_corner(target_number=1)
# fig2.write_image("corner.png",width=1800,height=1600)



import numpy as np 
import pypce
from collections import defaultdict
import time as T
import matplotlib.pyplot as mpl
import warnings, pdb
 
from sklearn.base import BaseEstimator, RegressorMixin, TransformerMixin
from sklearn.utils.estimator_checks import check_estimator 
from sklearn.utils.validation import check_X_y, check_array
from sklearn.model_selection import GridSearchCV, ParameterGrid, KFold, cross_val_score, cross_validate
from sklearn.ensemble import RandomForestRegressor
from sklearn.preprocessing import FunctionTransformer
from sklearn.metrics import make_scorer
from sklearn.decomposition import PCA, FastICA
		
class RegressionWrapperCV(BaseEstimator):
	def __init__(self,regressor='pce',reg_params={},scorer='neg_root_mean_squared_error',n_jobs=1,verbose=1,cv=None):
		self.regressor = regressor
		self.reg_params = reg_params
		self.scorer = scorer
		self.n_jobs = n_jobs
		self.verbose = verbose
		self.cv = cv
		# self.__dict__.update(reg_params)
	def _setup_cv(self):
		if self.cv == None:
			self.cv = KFold(n_splits=5)
	def _reformat_grid(self,params):
		try: 
			# if parameter grid is in the form of grid search cv
			ParameterGrid(params)
			params_cv = params
		except:
			# convert reg_params to form that can be used by grid search
			# for a single grid point
			params_cv = defaultdict(list)
			for key,item in params.items():
				params_cv[key].append(item)
		return params_cv
	def _model_factory(self,regressor):
		if regressor == 'pce':
			return pypce.PCEReg()
		if regressor == 'randforests':
			return RandomForestRegressor()
	def fit(self,X,y):
		self._setup_cv()
		if isinstance(self.regressor,str):
			model = self._model_factory(regressor=self.regressor)
			reg_params_cv = self._reformat_grid(self.reg_params)
			GridSCV = GridSearchCV(model, reg_params_cv, scoring=self.scorer, n_jobs=self.n_jobs, cv=self.cv, verbose=self.verbose, return_train_score=True)
			GridSCV.fit(X,y)
			self.best_estimator_ = GridSCV.best_estimator_
			self.best_params_ = GridSCV.best_params_
			self.best_score_ = GridSCV.best_score_
			self.best_overfit_error_ = self.overfit_score(GridSCV)
			self.cv_results_ = GridSCV.cv_results_
		if isinstance(self.regressor,list):
			self.fit_multiple_reg(X,y)
		return self
	def fit_multiple_reg(self,X,y):
		if isinstance(self.regressor,list):
			n_regressors = len(self.regressor)
			self.best_estimators_ = [None]*n_regressors
			self.all_best_params_ = [None]*n_regressors
			self.best_scores_ = np.zeros(n_regressors)
			self.best_overfit_errors_ = [None]*n_regressors
			self.all_cv_results_ = [None]*n_regressors
			assert isinstance(self.reg_params,list), "parameters must also be a list"
			assert len(self.reg_params) == n_regressors, "length of parameters and regressors must match."
			for i,R in enumerate(self.regressor):
				print("Fitting regressor ", R)
				model = self._model_factory(regressor=R)
				reg_params_cv = self._reformat_grid(self.reg_params[i])
				GridSCV = GridSearchCV(model, reg_params_cv, scoring=self.scorer, n_jobs=self.n_jobs, cv=self.cv, verbose=self.verbose, return_train_score=True)
				GridSCV.fit(X,y)
				self.best_estimators_[i] = GridSCV.best_estimator_
				self.all_best_params_[i] = GridSCV.best_params_
				self.best_scores_[i] = GridSCV.best_score_
				self.best_overfit_errors_[i] = self.overfit_score(GridSCV)
				self.all_cv_results_[i] = GridSCV.cv_results_
			self.best_index_ = np.argmax(self.best_scores_)
			self.best_estimator_ = self.best_estimators_[self.best_index_]
			self.best_params_ = self.all_best_params_[self.best_index_]
			self.best_score_ = self.best_scores_[self.best_index_]
			self.best_overfit_error_ = self.best_overfit_errors_[self.best_index_]
			self.cv_results_ = self.all_cv_results_[self.best_index_]
		return self
	def predict(self,X):
		return self.best_estimator_.predict(X)
	def overfit_score(self,GridSCV):
		''' Takes a grid search cv object and returns overfit score '''
		# assert hasattr(self,'GridSCV'), "Must run fit function."
		best_index_ = GridSCV.best_index_
		mean_train_score = GridSCV.cv_results_['mean_train_score']
		mean_test_score = GridSCV.cv_results_['mean_test_score']
		return mean_train_score[best_index_] - mean_test_score[best_index_]

from tqdm import tqdm
class MRegressionWrapperCV(BaseEstimator, RegressorMixin):
	def __init__(self, 
				regressor='pce', reg_params={'order':2},
				target_transform=None,
				target_transform_params={},
				scorer='neg_root_mean_squared_error',
				n_jobs=4,cv=None,
				mixed=True,
				verbose=1):
		self.regressor = regressor
		self.reg_params = reg_params
		self.scorer = scorer
		self.n_jobs = n_jobs
		self.cv = cv
		self.target_transform = target_transform
		self.target_transform_params = target_transform_params
		self.mixed = mixed # to control whether mixed surrogate for each target TBD
		self.verbose = verbose
	def _setupCV(self, shuffle=False, randstate=13):
		if self.cv == None:
			self.cv = KFold(n_splits=5)  # ,shuffle=True,random_state=13)
	def _fit_target_transform(self,Y):
		if self.target_transform is None:
			self.TT = FunctionTransformer(lambda Y: Y)
		else:
			self.TT = self.target_transform(**self.target_transform_params)
		self.TT.fit(Y)
		return self
	def fit(self,X,Y):
		self.n, self.dim = X.shape
		self._setupCV()
		self._fit_target_transform(Y)
		Yhat = self.TT.transform(Y)
		if Yhat.ndim == 1: Yhat = np.atleast_2d(Yhat).T
		assert len(Yhat) == self.n, "mistmatch in no of samples btwn X and Y."
		self.ntargets = Yhat.shape[1]
		self._setupCV()
		if isinstance(self.regressor,str):
			self.res = self.fit_single_reg(X,Yhat)
		elif isinstance(self.regressor,list) and self.mixed == False:
			self.res = self.fit_multiple_reg(X,Yhat)
		elif isinstance(self.regressor,list) and self.mixed == True:
			self.res = self.fit_single_reg(X,Yhat)
		return self
	def fit_single_reg(self, X, Y, regressor=None,reg_params=None):
		if regressor == None:
			regressor = self.regressor
			reg_params = self.reg_params
		res = defaultdict(list)
		for i in tqdm(range(self.ntargets)):
			reg = RegressionWrapperCV(
				regressor=regressor,reg_params=reg_params,
				n_jobs=self.n_jobs, scorer=self.scorer, cv=self.cv, verbose=self.verbose)
			reg.fit(X, Y[:, i])
			res['best_estimators_'].append(reg.best_estimator_)
			res['best_params_'].append(reg.best_params_)
			res['best_scores_'].append(reg.best_score_)
			res['best_overfit_error_'].append(reg.best_overfit_error_)
		self.__dict__.update(res)
		return res
	def fit_multiple_reg(self, X, Y):
		# wont execute if mixed is True
		if isinstance(self.regressor,list):
			mres = []
			for i,r in enumerate(self.regressor):
				res = self.fit_single_reg(X,Y,
								regressor=r,
								reg_params=self.reg_params[i])
				mres.append(res)
		return mres
	def predict(self,X):
		assert hasattr(self,'res'), "Must run fit."
		if isinstance(self.res,dict):
			Ypred = self._predict_single(X,res=self.res)
		elif isinstance(self.res,list):
			Ypred = self._predict_multiple(X,res=self.res)
		return Ypred
	def _predict_single(self,X,res=None):
		# assert isinstance(self.res,dict), "Must pass string as regressor OR list with mixed=True, otherwise, predict is ambiguous"
		assert isinstance(res,dict), "for single prediction, results must be a dictionary."
		Yhatpred_list = []
		for estimator in self.best_estimators_:
			if X.ndim == 1:
				X = np.atleast_2d(X)
			Yhatpred_list.append(estimator.predict(X))
		Yhatpred = np.array(Yhatpred_list).T
		Ypred = self.TT.inverse_transform(Yhatpred)
		return Ypred
	def _predict_multiple(self,X,res=None):
		# assert isinstance(self.res,dict), "Must pass string as regressor OR list with mixed=True, otherwise, predict is ambiguous"
		assert isinstance(res,list), "for multiple predictions, results must be a list of dictionaries."
		predictions = []
		for r in res:
			predictions.append(self._predict_single(X,r))
		return np.squeeze(np.array(predictions))
	def feature_importances_(self):
		assert hasattr(self,'res'), "Must run .fit() first!"
		FI_ = []
		for estimator in self.best_estimators_:
			fi = estimator.feature_importances_
			FI_.append(fi)
		return np.array(FI_)

class PCATargetTransform(BaseEstimator, TransformerMixin):
    def __init__(self, n_components=2, cutoff=1e-2, svd_solver='arpack'):
        self.n_components = n_components
        self.cutoff = cutoff
        self.svd_solver = svd_solver
        self.K = None
    def fit(self, Y):
        self.n, self.d = Y.shape
        if self.n_components == 'auto':
            self._compute_K(Y)
        if isinstance(self.n_components, int):
            self.K = self.n_components
            if self.n_components > self.d:
                warnings.warn("No of components greater than dimension. Setting to maximum allowable.")
                self.n_components = self.d
        assert self.K is not None, "K components is not defined or being set properly."
        self.pca = PCA(n_components=self.K, svd_solver=self.svd_solver)
        self.pca.fit(Y)
        self.cumulative_error = np.cumsum(self.pca.explained_variance_ratio_)
        return self
    def fit_transform(self, Y):
        self.fit(Y)
        return self.pca.transform(Y)
    def transform(self, Y):
        assert hasattr(self, 'pca'), "Perform fit first."
        return self.pca.transform(Y)
    def inverse_transform(self, Yhat):
        return self.pca.inverse_transform(Yhat)
    def _compute_K(self, Y, max_n_components=50):
        ''' automatically compute number of PCA terms '''
        pca = PCA(n_components=min(max_n_components, self.d), svd_solver=self.svd_solver)
        pca.fit(Y)
        cumulative_error = np.cumsum(pca.explained_variance_ratio_)
        # print(cumulative_error)
        # need to check whether to use + 1 or not
        loc = np.where(1 - cumulative_error <= self.cutoff)[0] + 1
        if loc.size == 0:
        	warnings.warn("Cutoff may be too big. Setting K = 16")
        	self.K = 16
        elif loc.size > 0:
	        self.K = loc[0]




import tesuract
import unittest
import numpy as np
import warnings
from sklearn.metrics import mean_squared_error
import os
from joblib import dump, load
import time
import pytest

relpath = tesuract.__file__[:-11]  # ignore the __init__.py specification
print("relpath:", relpath)


def mse(a, b):
    return mean_squared_error(a, b, squared=False)


@pytest.mark.unit
class TestPCEBuilderOnArbDomain(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.n, self.d = 7, 3
        self.rn = np.random.RandomState(23)
        self.X = 5 * self.rn.rand(self.n, self.d)
        self.ibounds = [(0, 5) for r in range(self.d)]

    def test_full_domain_scaling(self):
        n, d, X, ibounds = self.n, self.d, self.X, self.ibounds
        scaler = tesuract.preprocessing.DomainScaler(
            dim=d, input_range=ibounds, output_range=(-1, 1)
        )

        # print(X)
        X_scaled = scaler.fit_transform(X)
        pce = tesuract.PCEBuilder(order=4)
        Xhat_ref = pce.fit_transform(X_scaled)
        # X_scaled = scaler.transform(X)
        # print(X_scaled)
        scaler._range_check(X_scaled, [(-1, 1) for r in range(d)])
        scaler._range_check(X, [(0, 5) for r in range(d)])
        return Xhat_ref

    def test_pcebuild_raise_bound_assertion(self):
        n, d, X, ibounds = self.n, self.d, self.X, self.ibounds
        pce = tesuract.PCEBuilder(order=4)
        self.assertRaises(AssertionError, pce.fit_transform, X)

    def test_pcebuild_with_input_range(self):
        n, d, X, ibounds = self.n, self.d, self.X, self.ibounds
        pce = tesuract.PCEBuilder(order=4, input_range=ibounds)
        Xhat = pce.fit_transform(X)
        Xhat_ref = self.test_full_domain_scaling()
        assert (
            mse(Xhat_ref, Xhat) <= 1e-16
        ), "mismatch between expected poly transform and actual poly transform"

    def test_pcebuild_polyeval_with_input_range(self):
        n, d, X, ibounds = self.n, self.d, self.X, self.ibounds
        pce = tesuract.PCEBuilder(order=4, input_range=ibounds)
        pce.compile(X.shape[1])
        coef = np.zeros(35)
        coef[0] = 1.0
        ypred = pce.polyeval(X, coef)
        assert np.all(
            ypred == 1
        ), "vector must be all ones. Something may have gone wrong with polyeval and input bounds."

    def test_pcebuild_polyeval_assert_error_wo_input_range(self):
        n, d, X, ibounds = self.n, self.d, self.X, self.ibounds
        pce = tesuract.PCEBuilder(order=4)
        pce.compile(X.shape[1])
        coef = np.zeros(35)
        coef[0] = 1.0
        self.assertRaises(AssertionError, pce.polyeval, X, coef)


@pytest.mark.unit
class TestPCEBuilder(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.dim = 2
        self.rn = np.random.RandomState(23423)

    def test_pcebuilder_customM(self):
        p = tesuract.PCEBuilder(order=4)
        p.compile(dim=self.dim)
        M = p.mindex
        p2 = tesuract.PCEBuilder(customM=M)
        p2.compile(dim=self.dim)
        assert np.sum(M - p2.mindex) == 0

    def test_pcebuilder_customM_as_float(self):
        p = tesuract.PCEBuilder(order=4)
        p.compile(dim=self.dim)
        M = p.mindex
        M = M.astype("float")
        p2 = tesuract.PCEBuilder(customM=M)
        with warnings.catch_warnings(record=True) as w:
            p2.compile(dim=self.dim)
            assert (
                p2.mindex.dtype == "int"
            ), "Custom multiindex is not saved as an integer type."

    def test_fit_transform(self):
        p = tesuract.PCEBuilder(order=4)
        X = 2 * self.rn.rand(10, self.dim) - 1
        Xhat = p.fit_transform(X)
        assert (
            Xhat.shape[1] == p.mindex.shape[0]
        ), "Mistmatch between number of columns of Xhat (Vandermonde) and terms in multiindex."

    def test_fit_transform_w_normalized(self):
        p = tesuract.PCEBuilder(order=4, normalized=True)
        X = 2 * self.rn.rand(10, self.dim) - 1
        Xhat = p.fit_transform(X)
        int_P0 = 2 * np.mean(Xhat[:, 0])
        assert np.abs(int_P0 - 1.0) <= 1e-13, "First term should integrate to 1"

    def test_fit_transform_w_nonnormalized(self):
        p = tesuract.PCEBuilder(order=4, normalized=False)
        X = 2 * self.rn.rand(10, self.dim) - 1
        Xhat = p.fit_transform(X)
        int_P0 = 2 * np.mean(Xhat[:, 0])
        assert int_P0 == 2, "First term is not normalized and should integrate to 2"

    def test_fit_transform_single_input(self):
        p = tesuract.PCEBuilder(order=4)
        X = 2 * self.rn.rand(1, self.dim) - 1
        Xhat = p.fit_transform(X)
        Xhat = p.fit_transform(X[0])

    def test_pcebuilder_poleval_w_custom_coef(self):
        # define the mean solution as a PCE
        coef = np.array([0.0, -0.25, 1.0 / 3, np.sqrt(2)])  # ,-.05*np.sqrt(2)])
        pce = tesuract.PCEBuilder(order=3)
        pce.compile(dim=1)  # must compile to compute the multiindex
        polyfun = lambda x: pce.polyeval(x, c=coef)
        x = np.linspace(-1, 1, 10)
        y = polyfun(x)

    def test_polyeval_w_variable_coef_and_fixed_X(self):
        p = tesuract.PCEBuilder(order=3, store_phi=True)
        X = 2 * self.rn.rand(30, self.dim) - 1
        y = (
            3
            + 2 * 0.5 * (3 * X[:, 0] ** 2 - 1)
            + X[:, 1] * 0.5 * (3 * X[:, 0] ** 2 - 1)
        )
        c_true = np.array(
            [
                3.0,
                0.0,
                0.0,
                2.0,
                0.0,
                0.0,
                0.0,
                1.0,
                0.0,
                0.0,
            ]
        )
        Xhat = p.fit_transform(X)
        assert Xhat.shape[1] == len(c_true), "coefficient array size is incorrect."
        ypred = p.polyeval(c=c_true)
        assert np.sum(np.abs(ypred - y)) <= 1e-14, "poly eval is not converging."
        ctest = np.zeros(len(c_true))
        ctest[2] = 1.0
        ypred2 = p.polyeval(c=ctest)
        assert (
            np.sum(np.abs(ypred2 - X[:, 1])) <= 1e-15
        ), "polyeval with fixed X is not working right."

    def test_polyeval_w_fixed_coef_and_variable_X(self):
        c_true = np.array(
            [
                3.0,
                0.0,
                0.0,
                2.0,
                0.0,
                0.0,
                0.0,
                1.0,
                0.0,
                0.0,
            ]
        )
        X = 2 * self.rn.rand(30, self.dim) - 1
        y = (
            3
            + 2 * 0.5 * (3 * X[:, 0] ** 2 - 1)
            + X[:, 1] * 0.5 * (3 * X[:, 0] ** 2 - 1)
        )
        p = tesuract.PCEBuilder(order=3, coef=c_true)
        ypred = p.polyeval(X)
        assert mse(ypred, y) <= 1e-15, "fixed coef and variable X does not work."

    def test_polyeval_custom_M_and_custom_coef(self):
        p0 = tesuract.PCEBuilder(order=3)
        p0.compile(dim=2)
        M = p0.mindex[[0, 3, 7]]
        c_true = np.array(
            [
                3.0,
                0.0,
                0.0,
                2.0,
                0.0,
                0.0,
                0.0,
                1.0,
                0.0,
                0.0,
            ]
        )
        p = tesuract.PCEBuilder(customM=M, coef=c_true[[0, 3, 7]])
        X = 2 * self.rn.rand(30, self.dim) - 1
        y = (
            3
            + 2 * 0.5 * (3 * X[:, 0] ** 2 - 1)
            + X[:, 1] * 0.5 * (3 * X[:, 0] ** 2 - 1)
        )
        assert (
            mse(p.polyeval(X), y) <= 1e-15
        ), "polyeval not returning accurate prediction."

    def test_sobol_sensitivity_with_var_coef(self):
        p = tesuract.PCEBuilder(order=3)
        p.compile(dim=2)
        coef = np.array(
            [
                3.0,
                0.0,
                0.0,
                2.0,
                0.0,
                0.0,
                0.0,
                1.0,
                0.0,
                0.0,
            ]
        )
        s = p.computeSobol(c=coef)
        assert mse(s, [1.0, 0.07692307692307691]) == 0, "sobol indices are not right."

    def test_sobol_sensitivity_after_fit_transform(self):
        p = tesuract.PCEBuilder(order=3)
        X = 2 * self.rn.rand(30, self.dim) - 1
        p.fit_transform(X)
        coef = np.array(
            [
                3.0,
                0.0,
                0.0,
                2.0,
                0.0,
                0.0,
                0.0,
                1.0,
                0.0,
                0.0,
            ]
        )
        s = p.computeSobol(c=coef)
        assert mse(s, [1.0, 0.07692307692307691]) == 0, "sobol indices are not right."

    # def test_sobol_sensitivity_customM_and_coef_constructor(self):
    # 	p0 = tesuract.PCEBuilder(order=3)
    # 	p0.compile(dim=2)
    # 	M = p0.mindex[[0,3,7]]
    # 	coef = np.array([3.,2.,1.])
    # 	p = tesuract.PCEBuilder(customM=M,coef=coef)
    # 	# p.compile(dim=2)
    # 	s = p.computeSobol()
    # 	assert mse(s,[1.0, 0.07692307692307691]) == 0, "sobol indices are not right."
    def test_sobol_sensitivity_customM_and_coef_constructor(self):
        p0 = tesuract.PCEBuilder(order=3)
        p0.compile(dim=2)
        M = p0.mindex[[0, 3, 7]]
        coef = np.array([3.0, 2.0, 1.0])
        p = tesuract.PCEBuilder(customM=M, coef=coef)
        # p.compile(dim=2)
        s = p.computeSobol()
        assert mse(s, [1.0, 0.07692307692307691]) == 0, "sobol indices are not right."

    def test_sobol_sensitivity_w_normalization(self):
        p0 = tesuract.PCEBuilder(order=3)
        p0.compile(dim=2)
        M = p0.mindex[[0, 3, 7]]
        coef = np.array([3.0, 2.0, 1.0])
        p = tesuract.PCEBuilder(customM=M, coef=coef, normalized=True)
        # p.compile(dim=2)
        s = p.computeSobol()
        print(s, p.mindex)
        assert (
            mse(s, [1.0, 0.2]) == 0
        ), "sobol indices for normalized case is  not right."

    def test_1d_poly_feature_transform(self):
        N = 32  # number of x grid points
        x = np.linspace(-1, 1, N)[:, np.newaxis]
        rn = np.random.RandomState(23948)
        x_random = 2 * rn.rand(N)[:, np.newaxis] - 1
        x_chebychev = -np.cos(np.pi * x)

        # define test functions
        runge = lambda x: 1.0 / (1 + 16.0 * x**2)
        gibbs = lambda x: np.sign(np.sin(np.pi * (x - np.pi / 2)))

        # get polynomial features
        from sklearn.preprocessing import PolynomialFeatures

        P = 8
        polyfeatures = PolynomialFeatures(degree=P)
        X = polyfeatures.fit_transform(x)

        # Legendre features
        pce = tesuract.PCEBuilder(order=P)
        Xleg = pce.fit_transform(x)
        # print(Xleg.shape, Xleg)

    def test_memory_footprint_w_attr_assert(self):
        N = 2500  # number of x grid points
        rn = np.random.RandomState(9)
        X = 2 * rn.rand(N, 5) - 1.0

        # # define test functions
        # runge = lambda x: 1./(1 + 16.*np.sum(x,axis=1)**2)
        # gibbs = lambda x: np.sign(np.sin(np.pi*(np.sum(x,axis=1)-np.pi/2)))

        pceb = tesuract.PCEBuilder(order=4)  # store_phi=False default
        pceb.fit_transform(X)
        self.assertRaises(AttributeError, getattr, pceb, "Phi")

        pceb = tesuract.PCEBuilder(order=4, store_phi=True)
        pceb.fit_transform(X)
        assert pceb.Phi.shape[0] == N, "Phi is kept but is NOT the right shape. "

    def test_timing_for_repeated_transform(self):
        N = 2500  # number of x grid points
        rn = np.random.RandomState(9)
        X = 2 * rn.rand(N, 5) - 1.0

        pceb = tesuract.PCEBuilder(order=4)  # store_phi=False default
        start1 = time.time()
        pceb.fit_transform(X)
        end1 = time.time() - start1
        start2 = time.time()
        pceb.fit_transform(X)
        end2 = time.time() - start2
        # assert end2 < end1, "second fit transform should be faster since mindex is already computed."
        assert (
            pceb._mindex_compute_count_ == 1
        ), "mindex is being computed more than once or none at all."


# code to test the joint Sobol function in PCEBuilder
@pytest.mark.unit
class TestJointSobol(unittest.TestCase):
    def test_initialize_joint_sensitivity(self):
        p = tesuract.PCEBuilder(order=2)
        dim = 3
        p.compile(dim=dim)
        print(p.mindex)
        nterms, ndim = p.mindex.shape
        dummy_coef = np.ones(nterms)
        assert ndim == dim, "dimensions don't match"
        S2 = p.computeJointSobol(c=dummy_coef)
        print("\n", S2)

    def test_joint_sensitivity_2D_3rd(self):
        pce = tesuract.PCEReg(order=3)
        rn = np.random.RandomState(123)
        X = 2 * rn.rand(30, 2) - 1
        y = (
            3
            + 2 * 0.5 * (3 * X[:, 0] ** 2 - 1)
            + X[:, 1] * 0.5 * (3 * X[:, 0] ** 2 - 1)
        )
        pce.fit(X, y)
        coef_sum = np.sum(
            [
                pce.normsq[i] * pce.coef[i] ** 2
                for i in range(len(pce.coef))
                if ((pce.mindex[i, 0] > 0) and (pce.mindex[i, 1] > 0))
            ]
        )
        index = pce.joint_effects()[1, 0]
        var_temp = coef_sum / index
        var_tot = np.sum(
            [
                pce.normsq[i] * pce.coef[i] ** 2
                for i in range(len(pce.coef))
                if (np.sum(pce.mindex[i, :]) > 0)
            ]
        )
        assert (var_temp - var_tot) ** 2 < 1e-10, "total variance does not match"



import numpy as np
import pdb, warnings, pickle

def get_tuples(length, total):
    if length == 1:
        yield (total,)
        return

    for i in range(total + 1):
        for t in get_tuples(length - 1, total - i):
            yield (i,) + t


def multiindex2(dim, order):
    """Much faster total order construction"""
    L = []
    for o in range(order + 1):
        L = L + list([t[::-1] for t in get_tuples(dim, o)])
    return L


class RecursiveHypMultiIndex:
    """
    Adapted from MUQ (https://mituq.bitbucket.io/source/_site/index.html) but modified to
        be pure Python
    """

    def __init__(self, dim, order):
        self.maxOrder = order
        self.dim = dim
        self.minOrder = 0
        self.index = self.getIndex()

    def getIndex(self):
        output = []
        currDim = 0
        base = np.zeros(self.dim, dtype=int)
        # new for hyperbolic multiindex
        self.nugget = 1e-5
        q = 0.5
        maxNormPow = self.maxOrder**q + self.nugget
        output = self.RecHype(maxNormPow, output, currDim, base, q)
        return np.array(output, dtype=int)

    def RecHype(self, maxNormPow, output, currDim, base, q):
        total_order = MultiIndex(dim=self.dim, order=self.maxOrder)
        mindex = total_order.getIndex()
        mindex_new = mindex[
            [
                np.sum(m**q) ** (1.0 / q) < (self.maxOrder + self.nugget)
                for m in mindex
            ]
        ]
        return mindex_new


# multindex class
class MultiIndex:
    """
    Parts are adapted from UQTk (https://github.com/sandialabs/UQTk), but all code has been heavily modified to be pure Python.

    For total order multiindex construction
         Usage:
         M = MultiIndex(dim=2,order=2)
         print(M.getIndex())
         works with dim >= 1 and order >=1
    """

    def __init__(self, dim, order, mindex_type="total_order"):
        self.dim = dim
        self.order = order
        self.nPCTerms = self.computeNPCTerms()
        self.index_exists = False
        self.index = np.zeros((1, 1))
        if mindex_type == "total_order":
            self.getIndex()  # calculate multindex if not given as input
        if mindex_type == "hyperbolic":
            HmI = RecursiveHypMultiIndex(dim, order)
            self.index_exists = True  # index created flag
            self.index = HmI.index
            self.nPCTerms = len(HmI.index)

    def setIndex(self, mI):
        # to wrap custom array index as a index object
        self.index = mI
        self.nPCTerms = len(mI)

    def getIndex(self):
        if self.index_exists == False:
            # print('Index has not been set')
            # print('Setting to total order...')
            self.index = self.computeIndex()
            self.index_exists = True
        return self.index

    def computeNPCTerms(self):
        """Adapted form UQTk (https://github.com/sandialabs/UQTk)"""
        enume = 1
        denom = 1
        minNO = np.amin([self.order, self.dim])
        for k in range(minNO):
            enume *= self.order + self.dim - k
            denom *= k + 1
        nPCTerms = enume / denom
        return int(nPCTerms)

    def computeIndex(self):
        index = np.array(multiindex2(self.dim, self.order), dtype=int)
        return index



import numpy as np

# cross validation test
def multi_target_rel_l2_error(y_true,y_pred):
	if y_true.ndim == 2 and y_true.ndim == 2: 
	    error_array =  np.linalg.norm(y_true - y_pred,axis=1)/np.linalg.norm(y_true,axis=1)
	    return error_array.mean()
	else:
		return np.linalg.norm(y_true - y_pred)/np.linalg.norm(y_true)


def plot_samples(Y,x=None,ax=None,q=[.005,.5,.995],show_mean=True,label=None,color=None,xlabel='',ylabel='',fs=10,alpha=.2,rotation=None,labelpad=10):
	assert Y.ndim == 2, "Y must contain samples and have 2 dimensions."
	nsamples, dim = Y.shape
	mu = np.mean(Y,axis=0)
	var = np.var(Y,axis=0)
	Q = np.quantile(Y,q=q,axis=0)
	if ax is None:
		fig,ax = mpl.subplots()
	if x is None:
		x = list(np.arange(dim))
	q_label = r"{}".format(label + ' 99% Q')
	# mu_label = r"{}".format(label + ' median')
	mu_label = None
	if q is None:
		ax.fill_between(x, mu - np.sqrt(var), mu + np.sqrt(var),color=color, alpha=alpha,label=label)
		if show_mean:
			ax.plot(x, mu, '--',color=color, lw=1.5, alpha=0.5,label=mu_label)
	else:
		ax.fill_between(x, Q[0], Q[2],color=color, alpha=0.2,label=q_label)
		if show_mean:
			ax.plot(x, Q[1], '--',color=color, lw=1.5, alpha=0.5,label=mu_label)

	ax.set_xlabel(r"{}".format(xlabel),fontsize=fs)
	ax.set_ylabel(r"{}".format(ylabel),fontsize=fs,rotation=rotation,labelpad=labelpad)
	return ax


# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import sys, os

sys.path.append(os.path.abspath("."))
import doc

# -- Project information -----------------------------------------------------

project = "tesuract"
copyright = "2020, K. Chowdhary"
author = "K. Chowdhary"

# The full version, including alpha/beta/rc tags
release = "0.1.2"

# add path
sys.path.append(os.path.abspath("../../"))

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
# extensions = ['sphinx.ext.autodoc', 'sphinx.ext.autosummary',
#               'sphinx.ext.doctest',
#               'sphinx.ext.mathjax',
#               'sphinx.ext.viewcode',
#               'sphinx_rtd_theme',
#               'sphinx.ext.todo',
#               'numpydoc'
# ]
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "numpydoc",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.imgconverter",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
]

# napoleon_google_docstring = False
# napolean_numpy_docstring = True
todo_include_todos = True
# napoleon_include_init_with_doc = True # True = show init
# napoleon_use_admonition_for_examples = True
# napoleon_use_admonition_for_notes = False
# napoleon_use_ivar = False
# napoleon_use_param = True
numpydoc_class_members_toctree = True

autodoc_default_options = {
    "members": True,
    "inherited-members": True,
    "member-order": "bysource",
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
# pygments_style = 'monokai'
# html_theme = 'sphinx_rtd_theme'

# book theme
html_theme = "sphinx_book_theme"

html_theme_options = {
    "collapse_navigation": False,
    "repository_url": "https://github.com/kennychowdhary/tesuract",
    "use_repository_button": True,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

html_short_title = "tesuract"



import tesuract
import unittest
import numpy as np
import warnings
from sklearn.metrics import mean_squared_error
import pytest

relpath = tesuract.__file__[:-11]  # ignore the __init__.py specification
print(relpath)


def mse(a, b):
    return mean_squared_error(a, b, squared=False)


# Genz function tests
def genz_exp(x):
    if x.ndim == 2:
        output = np.exp(np.sum(x - 1, axis=1))
    elif x.ndim == 1:
        output = np.exp(np.sum(x - 1))
    return output


genz_exp_soln = lambda d: (np.exp(0) - np.exp(-1.0)) ** d


@pytest.mark.regression
class TestQuad(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.Q_d5_o5_LG_sp = tesuract.QuadGen(
            dim=5, order=5, rule="legendre_gauss", sparse=True
        )
        self.xw_d5_o5_LG_sp = self.Q_d5_o5_LG_sp.run()
        self.Q_d5_o5_LG_tp = tesuract.QuadGen(
            dim=5, order=5, rule="legendre_gauss", sparse=False
        )
        self.xw_d5_o5_LG_tp = self.Q_d5_o5_LG_tp.run()

        self.Q_d10_o2_LG_sp = tesuract.QuadGen(
            dim=10, order=2, rule="legendre_gauss", sparse=True
        )
        self.xw_d10_o2_LG_sp = self.Q_d10_o2_LG_sp.run()
        self.Q_d10_o2_LG_tp = tesuract.QuadGen(
            dim=10, order=2, rule="legendre_gauss", sparse=False
        )
        self.xw_d10_o2_LG_tp = self.Q_d10_o2_LG_tp.run()

        self.Q_d5_o5_CC_sp = tesuract.QuadGen(
            dim=5, order=5, rule="clenshaw_curtis", sparse=True
        )
        self.xw_d5_o5_CC_sp = self.Q_d5_o5_CC_sp.run()
        self.Q_d5_o5_CC_tp = tesuract.QuadGen(
            dim=5, order=5, rule="clenshaw_curtis", sparse=False
        )
        self.xw_d5_o5_CC_tp = self.Q_d5_o5_CC_sp.run()

        self.Q_d10_o2_CC_sp = tesuract.QuadGen(
            dim=10, order=2, rule="clenshaw_curtis", sparse=True
        )
        self.xw_d10_o2_CC_sp = self.Q_d10_o2_CC_sp.run()
        self.Q_d10_o2_CC_tp = tesuract.QuadGen(
            dim=10, order=2, rule="clenshaw_curtis", sparse=False
        )
        self.xw_d10_o2_CC_tp = self.Q_d10_o2_CC_tp.run()

    def test_sparse_quadrature_dim5_order5_LegGauss(self):
        # test...
        dim = 5
        # Q = tesuract.QuadGen(dim=dim,order=5,rule='legendre_gauss',sparse=True)
        x, w = self.xw_d5_o5_LG_sp
        approx = np.sum(genz_exp(x) * w)
        soln = genz_exp_soln(dim)
        error = np.mean((soln - approx) ** 2) / np.mean(soln**2)
        assert error <= 1e-15, "Quadrature error is too high. "

    def test_tensor_prod_quadrature_dim5_order5_LegGauss(self):
        # test...
        dim = 5
        # Q = tesuract.QuadGen(dim=dim,order=5,rule='legendre_gauss',sparse=False)
        x, w = self.xw_d5_o5_LG_tp
        approx = np.sum(genz_exp(x) * w)
        soln = genz_exp_soln(dim)
        error = np.mean((soln - approx) ** 2) / np.mean(soln**2)
        # print(soln, 'vs', approx,' , error = ', error)
        assert error <= 1e-30, "Quadrature error is too high. "

    def test_sparse_quadrature_dim5_order5_clenshaw_curtis(self):
        # test...
        dim = 5
        # Q = tesuract.QuadGen(dim=dim,order=5,rule='clenshaw_curtis',sparse=True)
        x, w = self.xw_d5_o5_CC_sp
        approx = np.sum(genz_exp(x) * w)
        soln = genz_exp_soln(dim)
        error = np.mean((soln - approx) ** 2) / np.mean(soln**2)
        # print(soln, 'vs', approx,' , error = ', error)
        assert error <= 5e-15, "Quadrature error is too high. "

    def test_sparse_vs_tensor_product_size_d10(self):
        # test...
        dim = 10
        # Q_sp = tesuract.QuadGen(dim=dim,order=2,rule='legendre_gauss',sparse=True)
        x_sp, w_sp = self.xw_d10_o2_LG_sp

        # Q_tp = tesuract.QuadGen(dim=dim,order=2,rule='legendre_gauss',sparse=False)
        x_tp, w_tp = self.xw_d10_o2_LG_tp

        assert (
            x_tp.shape[0] / x_sp.shape[0] >= 250.0
        ), "Sparse grid size should be much smaller than the full tensor product grid"

    def test_sparse_and_tensor_grid_dim_sizes(self):
        # test...
        dim = 10
        # Q_sp = tesuract.QuadGen(dim=dim,order=2,rule='legendre_gauss',sparse=True)
        x_sp, w_sp = self.xw_d10_o2_LG_sp

        # Q_tp = tesuract.QuadGen(dim=dim,order=2,rule='legendre_gauss',sparse=False)
        x_tp, w_tp = self.xw_d10_o2_LG_tp

        assert (
            x_tp.shape[1] == 10 and x_sp.shape[1] == 10
        ), "Dimensions of grids do not match input dimensions. "

    def test_sparse_and_tensor_grid_x_and_w_size_match_LG(self):
        # test...
        dim = 10
        # Q_sp = tesuract.QuadGen(dim=dim,order=2,rule='legendre_gauss',sparse=True)
        x_sp, w_sp = self.xw_d10_o2_LG_sp

        # Q_tp = tesuract.QuadGen(dim=dim,order=2,rule='legendre_gauss',sparse=False)
        x_tp, w_tp = self.xw_d10_o2_LG_tp

        assert x_tp.shape[0] == len(w_tp) and x_sp.shape[0] == len(
            w_sp
        ), "Grid and weight size mismatch."

    def test_sparse_and_tensor_grid_x_and_w_size_match_CC(self):
        # test...
        dim = 5
        # Q_sp = tesuract.QuadGen(dim=dim,order=5,rule='clenshaw_curtis',sparse=True)
        x_sp, w_sp = self.xw_d5_o5_CC_sp

        # Q_tp = tesuract.QuadGen(dim=dim,order=5,rule='clenshaw_curtis',sparse=False)
        x_tp, w_tp = self.xw_d5_o5_CC_tp

        assert x_tp.shape[0] == len(w_tp) and x_sp.shape[0] == len(
            w_sp
        ), "Grid and weight size mismatch."

    def test_sparse_and_tensor_weight_norm_check_to_1_d10_LG(self):
        # test...
        dim = 10
        # Q_sp = tesuract.QuadGen(dim=dim,order=2,rule='legendre_gauss',sparse=True)
        x_sp, w_sp = self.xw_d10_o2_LG_sp

        # Q_tp = tesuract.QuadGen(dim=dim,order=2,rule='legendre_gauss',sparse=False)
        x_tp, w_tp = self.xw_d10_o2_LG_tp

        assert (
            np.abs(np.sum(w_tp) - 1) <= 5e-14 and np.abs(np.sum(w_sp) - 1) <= 5e-14
        ), "Weights must sum to 1."

    def test_sparse_and_tensor_weight_norm_check_to_1_d5_LG(self):
        # test...
        dim = 5
        # Q_sp = tesuract.QuadGen(dim=dim,order=5,rule='legendre_gauss',sparse=True)
        x_sp, w_sp = self.xw_d5_o5_LG_sp

        # Q_tp = tesuract.QuadGen(dim=dim,order=5,rule='legendre_gauss',sparse=False)
        x_tp, w_tp = self.xw_d5_o5_LG_tp

        assert (
            np.abs(np.sum(w_tp) - 1) <= 5e-14 and np.abs(np.sum(w_sp) - 1) <= 5e-14
        ), "Weights must sum to 1."

    def test_sparse_and_tensor_weight_norm_check_to_1_d5_CC(self):
        # test...
        dim = 5
        # Q_sp = tesuract.QuadGen(dim=dim,order=5,rule='clenshaw_curtis',sparse=True)
        x_sp, w_sp = self.xw_d5_o5_CC_sp

        # Q_tp = tesuract.QuadGen(dim=dim,order=5,rule='clenshaw_curtis',sparse=False)
        x_tp, w_tp = self.xw_d5_o5_CC_tp

        assert (
            np.abs(np.sum(w_tp) - 1) <= 5e-14 and np.abs(np.sum(w_sp) - 1) <= 5e-14
        ), "Weights must sum to 1."

    def test_sparse_check_grid_domain_0_to_1_LG(self):
        # test...
        dim = 5
        # Q_sp = tesuract.QuadGen(dim=dim,order=5,rule='legendre_gauss',sparse=True)
        x_sp, w_sp = self.xw_d5_o5_LG_sp

        # Q_tp = tesuract.QuadGen(dim=dim,order=2,rule='legendre_gauss',sparse=False)
        x_tp, w_tp = self.xw_d5_o5_LG_tp

        assert np.all(x_sp >= 0.0) and np.all(
            x_sp <= 1.0
        ), "Sparse Grid points must be between 0 and 1. "

        assert np.all(x_tp >= 0.0) and np.all(
            x_tp <= 1.0
        ), "Tensor Product Grid points must be between 0 and 1. "

    def test_sparse_check_grid_domain_0_to_1_CC(self):
        # test...
        dim = 5
        # Q_sp = tesuract.QuadGen(dim=dim,order=5,rule='clenshaw_curtis',sparse=True)
        x_sp, w_sp = self.xw_d5_o5_CC_sp

        assert np.all(x_sp >= 0.0) and np.all(
            x_sp <= 1.0
        ), "Sparse Grid points must be between 0 and 1. "



import numpy as np 
import pypce
from collections import defaultdict

from sklearn.model_selection import GridSearchCV
from sklearn.base import BaseEstimator, RegressorMixin, TransformerMixin
from sklearn.utils.estimator_checks import check_estimator 
from sklearn.utils.validation import check_X_y, check_array
from sklearn.model_selection import KFold, cross_val_score, cross_validate
from sklearn.model_selection import ParameterGrid 
from sklearn.ensemble import RandomForestRegressor
from sklearn.preprocessing import FunctionTransformer

# data
n, d = 30, 2
rn = np.random.RandomState(45)
X = 2 * rn.rand(n, d) - 1
y = 3 + 2 * .5 * (3 * X[:, 0] ** 2 - 1) + X[:, 1] * .5 * (3 * X[:, 0] ** 2 - 1)
y2 = 1 + 2 * .5 * (3 * X[:, 1] ** 2 - 1) + 3 * X[:, 0] * .5 * (3 * X[:, 1] ** 2 - 1)
Y = np.vstack([y, y2]).T  # test for multioutput regressor
c_true = np.array([3., 0., 0., 2., 0., 0., 0., 1., 0., 0., ])


# domain test
class DomainScaler2:
	def __init__(self,dim,input_range,output_range=(-1,1)):
		self.dim = dim
		self.input_range = input_range
		self.output_range = output_range
	def _get_bound_list(self,input_range):
		if isinstance(input_range,list):
			assert len(input_range) == self.dim, "input range must be a list of tuples"
			input_bounds = input_range
		if isinstance(input_range,tuple):
			input_bounds = [(input_range[0],input_range[1]) for i in range(self.dim)]
		a = np.array([ab[0] for ab in input_bounds]) # lower bounds
		b = np.array([ab[1] for ab in input_bounds]) # upper bounds
		return input_bounds,a,b
	def _range_check(self,X,B):
		n,d = X.shape
		assert d == self.dim, "columns of X must be the same as dimensions"
		assert len(B) == self.dim, "length of bounds list must be same as dimensions"
		dim_check = [(X[:,i] >= B[i][0]).all() and (X[:,i] <= B[i][1]).all() for i in range(d)]
		assert all(dim_check), "X is not in the range of the input range."
		return X
	def fit(self,X):
		self.input_bounds,a,b = self._get_bound_list(self.input_range)
		self.output_bounds,c,d = self._get_bound_list(self.output_range)
		X = self._range_check(X,self.input_bounds)
		# transform to [0,1] first for ease
		X_unit_scaled = (X - a)/(b-a)
		# transform to output bounds
		X_scaled = (d-c)*X_unit_scaled + c
		X_scaled = self._range_check(X_scaled,self.output_bounds)
		return X_scaled
	def fit_transform(self,X):
		self.fit(X)
		return self
	def inverse_transform(self,Xhat):
		self.input_bounds,a,b = self._get_bound_list(self.input_range)
		self.output_bounds,c,d = self._get_bound_list(self.output_range)
		Xhat = self._range_check(Xhat,self.output_bounds)
		Xhat_unit_scaled = (Xhat - c)/(d-c)
		X_inv = (b-a)*Xhat_unit_scaled + a
		X_inv = self._range_check(X_inv,self.input_bounds)
		return X_inv 

X_1 = X.copy() # [-1,1] for both dimensions
X_2 = .5*(X+1) # [0,1] for both dimensions
X_3 = np.array([X.T[0]+1,3+.5*X.T[1]]).T # [0,2]x[2.5,3.5] 

D = DomainScaler2(dim=2,input_range=(-1,1),output_range=(-1,1))
X_scaled = D.fit(X_1) # difference should be zero
assert (X_scaled - X_1).sum() == 0

D = DomainScaler2(dim=2,input_range=(-1,1),output_range=(0,1))
X_scaled = D.fit(X_1) # difference should be zero
assert (np.amin(X_scaled,axis=0)>=0).all() and (np.amax(X_scaled,axis=0)<=1).all()

D = DomainScaler2(dim=2,input_range=(-1,1),output_range=[(0,1),(3,15)])
X_scaled = D.fit(X_1) # difference should be zero

D = DomainScaler2(dim=2,input_range=[(0,1),(3,15)],output_range=(-1,1))
X_scaled = D.fit(X_scaled) # difference should be zero

D = DomainScaler2(dim=2,input_range=(0,1),output_range=(-1,1))
X_scaled = D.fit(X_2) # difference should be zero

D = DomainScaler2(dim=2,input_range=[(0,2),(2.5,3.5)],output_range=(-1,1))
X_scaled = D.fit(X_3) # difference should be zero

D = DomainScaler2(dim=2,input_range=(-1,1),output_range=[(0,2),(2.5,3.5)])
X_scaled_new = D.fit(X_scaled) # difference should be zero
assert (X_scaled_new - X_3).sum() == 0

X_inv = D.inverse_transform(X_scaled_new)
assert (X_inv - X_scaled).sum() == 0



import tesuract
import unittest
import numpy as np
import warnings
from sklearn.metrics import mean_squared_error
import pytest

relpath = tesuract.__file__[:-11]  # ignore the __init__.py specification
print(relpath)


def mse(a, b):
    return mean_squared_error(a, b, squared=False)


@pytest.mark.unit
class TestPCERegressionArbDomain(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.dim = 2
        self.order = 3
        self.nsamples = 27
        rn = np.random.RandomState(123)
        X = 2 * rn.rand(self.nsamples, self.dim) - 1
        self.ibounds = [(-1, 1) for r in range(self.dim)]
        y = (
            3.14
            + 1.74 * 0.5 * (3 * X[:, 0] ** 2 - 1)
            + X[:, 1] * 0.5 * (3 * X[:, 0] ** 2 - 1)
        )
        self.c_true = np.array(
            [
                3.14,
                0.0,
                0.0,
                1.74,
                0.0,
                0.0,
                0.0,
                1.0,
                0.0,
                0.0,
            ]
        )
        self.X, self.y = X, y
        self.rn = rn

    def domain_setup(self, a=0, b=5):
        # on (0,5) domain
        X = (b - a) * self.rn.rand(self.nsamples, self.dim) + a
        ibounds = [(a, b) for r in range(self.dim)]
        y = (
            3
            + 2 * 0.5 * (3 * X[:, 0] ** 2 - 1)
            + X[:, 1] * 0.5 * (3 * X[:, 0] ** 2 - 1)
        )
        # self.c_true = np.array([3.14,0.,0.,1.74,0.,0.,0.,1.,0.,0.,])
        return X, y, ibounds

    def test_pcereg_w_canonical_domain(self):
        X, y, ibounds = self.X, self.y, self.ibounds
        pce = tesuract.PCEReg(order=3, input_range=ibounds)
        pce.fit(X, y)
        assert mse(y, pce.predict(X)) <= 1e-14
        # print(np.around(pce.coef_,3))
        assert mse(np.around(pce.coef_, 3), self.c_true) <= 1e-16

    def test_pcereg_w_non_unit_domain(self):
        X, y, ibounds = self.domain_setup()
        pce = tesuract.PCEReg(order=3, input_range=ibounds)
        pce.fit(X, y)
        assert (
            mse(y, pce.predict(X)) <= 1e-12
        ), "predict is not working properly for input ranges that are not on [-1,1]."

    def test_pcereg_w_unit_domain(self):
        X, y, ibounds = self.domain_setup(a=0, b=5)
        pce = tesuract.PCEReg(order=3, input_range=ibounds)
        pce.fit(X, y)
        assert (
            mse(y, pce.predict(X)) <= 1e-12
        ), "predict is not working properly for input ranges that are not on [-1,1]."


@pytest.mark.unit
class TestPCERegression(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.dim = 2
        self.order = 3
        self.nsamples = 30
        rn = np.random.RandomState(123)
        X = 2 * rn.rand(self.nsamples, self.dim) - 1
        y = (
            3
            + 2 * 0.5 * (3 * X[:, 0] ** 2 - 1)
            + X[:, 1] * 0.5 * (3 * X[:, 0] ** 2 - 1)
        )
        self.c_true = np.array(
            [
                3.0,
                0.0,
                0.0,
                2.0,
                0.0,
                0.0,
                0.0,
                1.0,
                0.0,
                0.0,
            ]
        )
        self.X, self.y = X, y
        self.rn = rn

    def test_linear_fit(self):
        p = tesuract.PCEReg(self.order)
        p.fit(self.X, self.y)
        assert (
            np.mean(np.abs(p.coef - self.c_true)) <= 1e-12
        ), "Fit coefficient did not match truth."

    def test_linear_fit_w_normalization(self):
        p = tesuract.PCEReg(self.order, normalized=True)
        p.fit(self.X, self.y)
        normsq = p.computeNormSq()
        assert (
            mse(p.coef / np.sqrt(normsq), self.c_true) <= 5e-12
        ), "Fit coefficient did not match truth."

    def test_linear_fit_prediction_w_normalization(self):
        p = tesuract.PCEReg(self.order, normalized=True)
        p.fit(self.X, self.y)
        assert (
            mse(p.predict(self.X), self.y) <= 5e-12
        ), "linear fit with normalization is broken."

    def test_multindex(self):
        p = tesuract.PCEReg(self.order)
        p.fit(self.X, self.y)
        M = p._M
        assert np.sum(M - p.mindex) == 0, "Multindex does not match."

    def test_custom_index(self):
        p0 = tesuract.PCEReg(self.order)
        p0.fit(self.X, self.y)
        p = tesuract.PCEReg(order=self.order, customM=p0._M)
        p.fit(self.X, self.y)
        assert (
            np.mean(np.abs(p.coef - self.c_true)) <= 1e-12
        ), "Fit coefficient did not match truth."

    def test_custom_feature_importance(self):
        rn = np.random.RandomState(123)
        X = 2 * rn.rand(100, 2) - 1
        y = X[:, 0] + 0.5 * (3 * X[:, 1] ** 2 - 1)
        p0 = tesuract.PCEReg(self.order)
        p0.fit(X, y)
        fi0 = p0.feature_importances_
        p = tesuract.PCEReg(customM=p0.mindex[1:])
        p.fit(X, y)
        fi = p.feature_importances_
        assert (
            np.sum((fi - fi0) ** 2) <= 1e-16
        ), "feature importance for custom multiindex failed."
        # assert np.mean(np.abs(p.coef - self.c_true)) <= 1e-12, "Fit coefficient did not match truth."

    def test_LassoCV_fit(self):
        p = tesuract.PCEReg(self.order, fit_type="LassoCV")
        y = self.y + 0.001 * self.rn.rand(len(self.y))  # add noise
        p.fit(self.X, y)
        assert (
            np.mean(np.abs(p.coef - self.c_true)) <= 2e-2
        ), "LassoCV not accurate enough."

    def test_OmpCV_fit(self):
        p = tesuract.PCEReg(self.order, fit_type="OmpCV")
        y = self.y + 0.001 * self.rn.rand(len(self.y))  # add noise
        p.fit(self.X, y)
        assert (
            np.mean(np.abs(p.coef - self.c_true)) <= 2e-1
        ), "OmpCV not accurate enough."

    def test_ElasticNetCV_fit(self):
        p = tesuract.PCEReg(self.order, fit_type="ElasticNetCV")
        p.fit(self.X, self.y)
        assert (
            np.mean(np.abs(p.coef - self.c_true)) <= 2e-2
        ), "ElasticNetCV error is not small enough."

    def test_predict_in_base_vs_child_class(self):
        p = tesuract.PCEReg(self.order)
        p.fit(self.X, self.y)
        ypred1 = p.predict(self.X)
        ypred2 = p.polyeval(self.X)
        assert (
            np.mean(np.abs(ypred1 - ypred2)) == 0
        ), "Mismatch between eval and prediction."

    def test_predict_in_base_vs_child_with_normalization(self):
        p = tesuract.PCEReg(self.order, normalized=True)
        p.fit(self.X, self.y)
        ypred1 = p.predict(self.X)
        ypred2 = p.polyeval(self.X)
        assert (
            np.mean(np.abs(ypred1 - ypred2)) == 0
        ), "Mismatch between eval and prediction."

    def test_single_predict(self):
        p = tesuract.PCEReg(self.order)
        p.fit(self.X, self.y)
        ypred1 = p.predict(self.X[0])
        ypred2 = p.polyeval(self.X[0])
        assert (
            np.mean(np.abs(ypred1 - ypred2)) == 0
        ), "Mismatch between eval and prediction."

    def test_set_custom_multiindex(self):
        M = np.array([[0, 0], [2, 0], [2, 1]])
        p = tesuract.PCEReg(order=self.order, customM=M)
        p.fit(self.X, self.y)
        assert (
            np.mean(p.coef - np.array([3.0, 2.0, 1.0])) <= 1e-10
        ), "coefficients did not converge for custom multindex."
        assert (
            np.mean(np.abs(self.y - p.predict(self.X))) <= 1e-10
        ), "Fit did not converge and/or predict did not match pce.eval."

    def test_grid_search_cv(self):
        X = 2 * self.rn.rand(self.nsamples, self.dim) - 1
        y = (
            3
            + 2 * 0.5 * (3 * X[:, 0] ** 2 - 1)
            + X[:, 1] * 0.5 * (3 * X[:, 0] ** 2 - 1)
        )
        param_grid = [
            {
                "order": [1, 3],
                "mindex_type": [
                    "total_order",
                    "hyperbolic",
                ],
                "fit_type": ["LassoCV", "linear"],
            }
        ]
        from sklearn.model_selection import GridSearchCV

        pceCV = GridSearchCV(
            tesuract.PCEReg(), param_grid, scoring="neg_root_mean_squared_error"
        )
        pceCV.fit(X, y)
        assert mse(self.c_true.shape, pceCV.best_estimator_.coef.shape) < 1e-12

    def test_fit_with_2d_quadrature(self):
        X = np.loadtxt(relpath + "/tests/data/X_2dquad_points.txt")
        y = (
            3
            + 2 * 0.5 * (3 * X[:, 0] ** 2 - 1)
            + X[:, 1] * 0.5 * (3 * X[:, 0] ** 2 - 1)
        )
        c0 = np.array(
            [
                3.0,
                0.0,
                0.0,
                2.0,
                0.0,
                0.0,
                0.0,
                1.0,
                0.0,
                0.0,
            ]
        )
        w = np.loadtxt(relpath + "/tests/data/w_2dquad_weights.txt")
        p = tesuract.PCEReg(
            order=3, normalized=False, fit_type="quadrature", fit_params={"w": w}
        )
        p.fit(X, y)
        assert (
            mse(p.coef, c0) <= 1e-14
        ), "quadratue with non normalized basis did not work. "

    def test_fit_with_2d_quadrature_and_normalized_basis(self):
        X = np.loadtxt(relpath + "/tests/data/X_2dquad_points.txt")
        y = (
            3
            + 2 * 0.5 * (3 * X[:, 0] ** 2 - 1)
            + X[:, 1] * 0.5 * (3 * X[:, 0] ** 2 - 1)
        )
        c0 = np.array(
            [
                3.0,
                0.0,
                0.0,
                2.0,
                0.0,
                0.0,
                0.0,
                1.0,
                0.0,
                0.0,
            ]
        )
        w = np.loadtxt(relpath + "/tests/data/w_2dquad_weights.txt")
        p = tesuract.PCEReg(
            order=3, normalized=False, fit_type="quadrature", fit_params={"w": w}
        )
        p.fit(X, y)
        assert (
            mse(p.coef, c0) <= 1e-14
        ), "quadratue with non normalized basis did not work. "

    def test_quadrature_assert_error_wo_no_weights(self):
        X = np.loadtxt(relpath + "/tests/data/X_2dquad_points.txt")
        y = (
            3
            + 2 * 0.5 * (3 * X[:, 0] ** 2 - 1)
            + X[:, 1] * 0.5 * (3 * X[:, 0] ** 2 - 1)
        )
        c0 = np.array(
            [
                3.0,
                0.0,
                0.0,
                2.0,
                0.0,
                0.0,
                0.0,
                1.0,
                0.0,
                0.0,
            ]
        )
        w = np.loadtxt(relpath + "/tests/data/w_2dquad_weights.txt")
        p = tesuract.PCEReg(order=3, normalized=False, fit_type="quadrature")
        with self.assertRaises(AssertionError):
            p.fit(X, y)

    def test_fit_predict_w_store_false(self):
        p2 = tesuract.PCEReg(self.order, fit_type="linear")
        y2 = self.y + 0 * 0.001 * self.rn.rand(self.nsamples)  # add noise
        p2.fit(self.X, y2)
        # print(np.around(p2.coef_))
        assert mse(p2.predict(self.X), y2) < 1e-12, "accuracy loss. check. "
        self.assertRaises(AttributeError, getattr, p2, "Phi")

    def test_fit_w_store_true(self):
        p2 = tesuract.PCEReg(self.order, fit_type="linear", store_phi=True)
        y2 = self.y + 0 * 0.001 * self.rn.rand(self.nsamples)  # add noise
        p2.fit(self.X, y2)
        assert hasattr(
            p2, "Phi"
        ), "object does not have Phi, but it should if store_phi=True"

    def test_Xhat_w_store_true(self):
        p2 = tesuract.PCEReg(self.order, fit_type="linear", store_phi=True)
        y2 = self.y + 0 * 0.001 * self.rn.rand(self.nsamples)  # add noise
        p2.fit(self.X, y2)
        assert mse(p2.predict(self.X), y2) < 1e-12, "accuracy loss. check. "
        assert hasattr(
            p2, "Xhat"
        ), "object does not have Xhat, but it should if store_phi=True"

    def test_Xhat_w_store_false(self):
        p2 = tesuract.PCEReg(self.order, fit_type="linear")
        y2 = self.y + 0 * 0.001 * self.rn.rand(self.nsamples)  # add noise
        p2.fit(self.X, y2)
        assert mse(p2.predict(self.X), y2) < 1e-12, "accuracy loss. check. "
        self.assertRaises(AttributeError, getattr, p2, "Xhat")



import numpy as np
import matplotlib.pyplot as mpl
import pdb, warnings, pickle
from tqdm import tqdm
from sklearn.decomposition import PCA
from sklearn.base import BaseEstimator, RegressorMixin, TransformerMixin
from sklearn.preprocessing import MinMaxScaler

from .multiindex import RecursiveHypMultiIndex
from .multiindex import MultiIndex

from sklearn.utils.estimator_checks import check_estimator 
from sklearn.utils.validation import check_X_y, check_array
from sklearn import linear_model

# Polynomials classes
class PolyBase:
	def __init__(self,K):
		pass
	def Eval1dBasis(self,x):
		# virtual fun to compute basis values at x up to order K
		pass
	def Eval1dBasisNorm(self,x):
		# virtual fun to compute norm basis values at x up to order K
		pass

from numpy.polynomial.legendre import Legendre
class LegPoly(PolyBase):
	'''
	Usage for 1d polynomial basis construction
	L = LegPoly()
	x = np.linspace(-1,1,100)
	L.plotBasis(x,K,normed=True)
	'''
	def __init__(self,domain=np.array([-1,1])):
		self.domain = domain
	def Eval1dBasis(self,x,K):
		# returns matrix where each column is Li(x)
		self.K = K
		self.Lis = [Legendre.basis(k,self.domain) for k in range(self.K+1)]
		return np.array([Li(x) for Li in self.Lis]).T
	def Eval1dBasisNorm(self,x,K):
		# returns matrix where each column is Li(x)
		self.K = K
		self.Lis = [Legendre.basis(k,self.domain) for k in range(self.K+1)]
		return np.array([Li(x)/self.normsq[i] for i,Li in enumerate(self.Lis)]).T
	def normsq(self,K):
		# compute the squared norm of each basis up to K
		bma = self.domain[1] - self.domain[0]
		normsq = np.array([bma/(2*k+1) for k in range(K+1)])
		self.normsq = normsq
		return normsq
	def plotBasis(self,x,K,normed=False):
		if normed == False: 
			output = self.Eval1dBasis(x,K)
		elif normed == True:
			output = self.Eval1dBasisNorm(x,K)
		[mpl.plot(Li) for Li in output]
		mpl.show()

class PolyFactory:
	# generates PolyBase class object
	@staticmethod
	def newPoly(polytype='Legendre'):
		L = LegPoly()
		return L

class scale:
	def __init__(self,X,a=None,b=None):
		self.X = np.atleast_2d(X)
		self.dim = self.X.shape[1] # each column is a dimension or feature
		if a is None and b is None:
			# default bounds
			self.a = -1*np.ones(self.dim)
			self.b =  1*np.ones(self.dim)
			self.__scaled = self.X # already scaled
		elif len(a)==0 and len(b)==0:
			self.a = -1*np.ones(self.dim)
			self.b =  1*np.ones(self.dim)
			self.__scaled = self.X # already scaled
		else:
			self.a = a
			self.b = b
			assert self.dim == len(a), "dim mismatch in a!"
			assert self.dim == len(b), "dim mismatch in b!"
			
			self.w = self.b - self.a
			assert self.X.shape[1] == len(self.w),"X and a,b mismatch!"
			# scale to -1 -> 1
			self.__scaled = 2*(self.X - self.a)/self.w - 1.0
		# integration factor ([a,b] -> [-1,1])
		# self.intf = np.prod(.5*(self.b - self.a))
		self.intf = np.prod(.5*(self.b - self.a)/(self.b - self.a)) # (b-a) canceled by prod prob weight
	def __getitem__(self,key):
		return self.__scaled[key]

class DomainScaler:
	def __init__(self,dim=None,a=None,b=None,domain_range=(-1,1)):
		if a is None:
			assert dim is not None, "Must defined dimension if a or b is None."
			self.dim = dim
		else:
			self.dim = len(a)
			assert len(a) == len(b), "Mismatch between a and b"
		self.a = a
		self.b = b
		self.dr_a,self.dr_b = domain_range
	def _compile(self):
		self.dr_w = self.dr_b - self.dr_a
		if self.a is None and self.b is None:
			# default bounds
			self.a = self.dr_a*np.ones(self.dim)
			self.b = self.dr_b*np.ones(self.dim)
		else:
			self.w = self.b - self.a
		# integration factor ([a,b] -> [-1,1])
		# self.intf = np.prod(.5*(self.b - self.a))
		self.intf = np.prod((1./self.dr_w)*(self.b - self.a)/(self.b - self.a)) # (b-a) canceled by prod prob weight
	def fit(self,X):
		self._compile() # get domain width
		X = np.atleast_2d(X)
		assert self.dim == X.shape[1], "Size of data matrix features does not match dimensions." 
		# scale to -1 -> 1
		# self.X_scaled_ = 2*(self.X - self.a)/self.w - 1.0
		return self
	def transform(self,X):
		X = np.atleast_2d(X)
		assert self.dim == X.shape[1], "Size of data matrix features does not match dimensions." 
		# scale to -1 -> 1
		X_scaled_ = self.dr_w*(X - self.a)/self.w + self.dr_a
		return X_scaled_
	def fit_transform(self,X):
		self.fit(X)
		return self.transform(X)
	def inverse_transform(self,Xhat):
		Xhat = np.atleast_2d(Xhat)
		assert self.dim == Xhat.shape[1], "Size of data matrix features does not match dimensions." 
		X = (1./self.dr_w)*(Xhat - self.dr_a) # [-1,1] -> [0,1]
		X = self.a + self.w*X
		return X

class MinMaxTargetScaler:
	'''
	Scales the entire target matrix with single scalar min and max. This way it is easy to invert for new predicted values.
	'''
	def __init__(self,target_range=(0,1)):
		self.a,self.b = target_range
	def fit(self,Y):
		self.min_ = np.amin(Y)
		self.max_ = np.amax(Y)
		self.w_ = self.max_ - self.min_
		self.ab_w_ = self.b - self.a
	def transform(self,Y):
		assert self.min_ is not None and self.max_ is not None, "Need to fit first."
		# scale to 0 -> 1 first
		Yhat = (Y - self.min_)/self.w_
		# scale to target_range
		Yhat = (Yhat - self.a)/self.ab_w_ 
		return Yhat
	def fit_transform(self,Y):
		self.fit(Y)
		return self.transform(Y)
	def inverse_transform(self,Yhat):
		# first transform back to [0,1]
		Y = Yhat*self.ab_w_ + self.a
		# back to true range
		Y = self.w_*Y  + self.min_
		return Y

class PCEBuilder(BaseEstimator):
	def __init__(self,order=1,customM=None,mindex_type='total_order',coef=None,a=None,b=None,polytype='Legendre'):
		# self.dim = dim # no need to initialize with dim
		self.order = order
		self.dim = None
		self.customM = customM
		self.mindex_type = mindex_type
		self.coef = coef
		# self.coef_ = self.coef
		self.polytype = polytype
		self.a = a # lower bounds on domain of x
		self.b = b # upper bound on domain of x
		self.normsq = np.array([])
		self.compile_flag = False
		self.mindex = None
	def compile(self,dim=None):
		self.dim = dim
		#constructor for different multindices (None, object, and array)
		if self.customM is None:
			self.M = MultiIndex(dim,self.order,self.mindex_type)
		elif isinstance(self.customM,MultiIndex):
			# print("Using custom Multiindex object.")
			assert(self.customM.dim == dim)
			self.M = customM
		elif isinstance(self.customM,np.ndarray):
			self.dim = self.customM.shape[1] # set dim to match multiindex
			self.M = MultiIndex(self.dim,order=1,mindex_type='total_order') # use default
			self.M.setIndex(self.customM)
			self.order = None # leave blank to indicate custom order
		self.mindex = self.M.index
		if self.mindex.dtype != 'int':
			warnings.warn("Converting multindex array to integer array.")
			self.mindex = self.mindex.astype('int')
		self.nPCTerms = self.mindex.shape[0]
		if self.coef is not None: 
			assert len(self.coef) == self.mindex.shape[0], "coefficient array is not the same size as the multindex array."
		self.compile_flag = True
	def computeNormSq(self):
		mindex = self.mindex # use internal mindex array
		# compute multindex
		Kmax = np.amax(mindex,0)
		NormSq = []
		for d in range(self.dim):
			# Compute Legendre objects and eval 1d basis
			# Future work: replace this with Hermite
			# 	and polybase/poly factory class in order to 
			# 	generate mixed representations
			# 	E.g.: L = PolyFactory(PolyObj) for each dim
			#   will give a different poly object 
			assert self.polytype == 'Legendre', "Only works with Legendre for now!"
			L = PolyFactory.newPoly(self.polytype)
			NormSq.append(L.normsq(Kmax[d])) # norms up to K order
		# Multiply basis to eval nd polynomial
		normsq = NormSq[0][mindex[:,0]]
		# print(norms)
		for n in range(1,self.dim):
			norms_new = NormSq[n][mindex[:,n]]
			normsq = np.multiply(normsq,norms_new) # get norm squared of basis functions
		self.normsq = .5*normsq # norm squared * integration factor of 1/2 (since domain is [-1,1])
		return self.normsq
	def fit_transform(self,X):
		# compute multindex
		X = np.atleast_2d(X)
		if self.mindex is None:
			self.compile(dim=X.shape[1]) # only compiles once
		# evaluate basis at X which is n x dim
		# first compute max order for each dimension
		# basis is constructed based on multiindex
		if self.a is not None or self.b is not None: 
			assert len(self.a)==self.dim and len(self.b)==self.dim, "For user-specified bounds, they must be the same size as dimension."
		Xs = scale(X,a=self.a,b=self.b)  # scale to [-1,1]
		intf = Xs.intf
		self.intf = intf # save integration factor for ref
		X = np.atleast_2d(Xs[:]) # in case len(X) = 1
		Kmax = np.amax(self.mindex,0)
		P = []
		Px = []
		NormSq = []
		for d in range(self.dim):
			# Compute Legendre objects and eval 1d basis
			# Future work: replace this with Hermite
			# 	and polybase/poly factory class in order to 
			# 	generate mixed representations
			# 	E.g.: L = PolyFactory(PolyObj) for each dim
			#   will give a different poly object 
			assert self.polytype == 'Legendre', "Only works with Legendre for now!"
			L = PolyFactory.newPoly(self.polytype)
			P.append(L)
			Lix = L.Eval1dBasis(X[:,d],Kmax[d]) # each dimension may use different points so we need to loop
			# print(Lix.shape)
			Px.append(Lix) # list of size dim of Legendre polys
			NormSq.append(L.normsq(Kmax[d])) # norms up to K order
		# Multiply basis to eval nd polynomial
		Index = self.mindex
		Phi = Px[0] # get first group of Legendre basis polynomials
		Phi = Phi[:,Index[:,0]] # reorder according to multi-index for first dimension
		normsq = NormSq[0][Index[:,0]]
		# print(norms)
		for n in range(1,self.dim):
			Phi_temp = Px[n] # get next group of Leg basis polynomaisl
			Phi_new = Phi_temp[:,Index[:,n]] # order according to next multindex
			Phi = np.multiply(Phi,Phi_new) # multiple previous matrix
			norms_new = NormSq[n][Index[:,n]]
			normsq = np.multiply(normsq,norms_new) # get norm squared of basis functions
			# print(norms_new)

		# internally set Phi and norms for repeated use in eval
		self.Phi = Phi
		self.normsq = .5*normsq # norm squared
		return self.Phi
	def polyeval(self,X=None,c=None):
		if c is not None: 
			coef_ = c # use c in polyeval
			assert self.compile_flag == True, "Must compile to get multindex."
			assert len(coef_) == self.mindex.shape[0], "c is not the right size."
		elif c is None: 
			assert self.coef is not None, "Must define coefficient in constructor or polyeval"
			coef_ = self.coef # use coefficient from constructor

		if X is not None:
			Phi = self.fit_transform(X)
			yeval = np.dot(Phi,coef_)
		elif X is None:
			Phi = self.Phi # using stored Vandermonde matrix
			yeval = np.dot(Phi,coef_)

		return yeval
	def eval(self,X=None,c=None):
		return self.polyeval(X=None,c=None)
	def computeSobol(self,c=None):
		if self.mindex is None:
			self.compile(dim=self.dim)
		# assert self.compile_flag == True, "Must compile to set mindex. Avoids having to recompute multiindex everytime we run SA."
		if len(self.normsq) == 0: 
			normsq = self.computeNormSq()
		else:
			normsq = self.normsq # from fit transform
		# compute Sobol indices using coefficient vector
		if c is None:
			assert len(self.coef) != 0, "must specify coefficient array or feed it in the constructor."
			coef_ = self.coef
		else:
			coef_ = c

		assert len(coef_) == self.mindex.shape[0], "Coefficient vector must match the no of rows of the multiindex."

		totvar_vec = normsq[1:]*coef_[1:]**2
		totvar = np.sum(totvar_vec)
		# assert totvar > 0, "Coefficients are all zero!"
		S = []
		for i in range(self.dim):
			si = self.mindex[:,i] > 0 # boolean
			s = np.sum(totvar_vec[si[1:]])/totvar
			S.append(s)
		# in case elements have nan in them
		if np.all(coef_[1:] == 0): # ignore the mean
			print("Returning equal weights!")
			S = np.ones(self.dim)/self.dim # return equal weights
		return S

class pcereg(PCEBuilder,RegressorMixin):
	def __init__(self,order=2,customM=None,mindex_type='total_order',coef=None,a=None,b=None,polytype='Legendre',fit_type='linear',alphas=np.logspace(-12,1,20),l1_ratio=[.001,.5,.75,.95,.999,1],lasso_tol=1e-2):
		self.order = order
		self.mindex_type = mindex_type
		self.customM = customM
		self.a = a
		self.b = b
		self.coef = coef
		# self.c = self.coef # need to remove eventually
		self.polytype = 'Legendre'
		self.fit_type = fit_type
		self.alphas = alphas # LassoCV default
		self.l1_ratio = l1_ratio # ElasticNet default
		self.lasso_tol = lasso_tol
		super().__init__(order=order,customM=customM,mindex_type=mindex_type,coef=coef,a=a,b=b,polytype=polytype)
	def compile(self,X):
		# build multindex
		self._dim = X.shape[1]
		super().compile(dim=self._dim) # use parent compile to produce the multiindex
		self._M = self.mindex
		self.Xhat = self.fit_transform(X)
		return self
	def fit(self,X,y):
		X,y = check_X_y(X,y)
		# get data attributes
		self._n,self._dim = X.shape
		self.compile(X) # build multindex and construct basis
		# pypce.PCEBuilder(dim=self.dim,self.order)
		if self.coef is not None:
			assert len(self.coef) == self._M.shape[0],"length of coefficient vector is not the same shape as the multindex!"
		else:
			# LINEAR fit
			Xhat,y = check_X_y(self.Xhat,y)
			if self.fit_type == 'linear':
				regmodel = linear_model.LinearRegression(fit_intercept=False)
				regmodel.fit(Xhat,y)
			elif self.fit_type == 'LassoCV':
				regmodel = linear_model.LassoCV(alphas=self.alphas,fit_intercept=False,max_iter=1000,tol=self.lasso_tol)
				regmodel.fit(Xhat,y)
				self.alpha_ = regmodel.alpha_
			elif self.fit_type == 'OmpCV':
				regmodel = linear_model.OrthogonalMatchingPursuitCV(fit_intercept=False)
				regmodel.fit(Xhat,y)
			elif self.fit_type == 'ElasticNetCV':
				regmodel = linear_model.ElasticNetCV(l1_ratio=self.l1_ratio,fit_intercept=False,n_alphas=25,tol=1e-2)
				regmodel.fit(Xhat,y)
			self.coef = regmodel.coef_
			# self.pce.coef_ = self.coef_ # set internal pce coef
		return self
	def sensitivity_indices(self):
		assert self.coef is not None, "Must run fit or feed in coef array."
		return self.computeSobol()
	def multiindex(self):
		assert self._M is not None, "Must run fit or feed in mindex array."
		return self._M
	def predict(self,X):
		# check_is_fitted(self)
		X = np.atleast_2d(X)
		X = check_array(X)
		# self._fit_transform(X)
		# Phi = check_array(self._Phi)
		# ypred = np.dot(Phi,self.coef_)
		ypred = self.polyeval(X)
		return ypred





# In development
class pceseries():
	def __init__(self,dim,C,MIs,time=[],a=None,b=None):
		# construct series of pce's
		assert len(C) == len(MIs), 'coef and multindex mismatch!'
		# add assert for single array
		if len(time) == 0:
			# assign simple index as time
			time = np.arange(len(C))
		self.time = time
		self.C = C
		self.MIs = MIs
		self.N = len(C)
		self.dim = dim
		self.a = a
		self.b = b
		# construct PCEs
		self.__constructPCEs()
	def __constructPCEs(self):
		self.PCEs = []
		for i in range(self.N):
			pcetemp = PCEBuilder(dim=self.dim,
								 customM=self.MIs[i],
								 c=self.C[i],a=self.a,b=self.b)
			self.PCEs.append(pcetemp)
	def eval(self,X):
		assert np.atleast_2d(X).shape[1] == self.dim, "dimension mismatch!"
		y = []
		for i in range(self.N):
			ytemp = self.PCEs[i].eval(X)
			y.append(ytemp)
		y = np.array(y).T
		return y
	def ieval(self,T,X):
		# each element of T can be a vector
		# first evaluate all PCEs at X points
		Y = self.eval(X)
		# interpolate for each time point
		T1d = np.atleast_1d(T)
		assert len(Y) == len(T1d), "time and X mismatch!"
		Yi = []
		for i in range(len(T1d)):
			yitemp = np.interp(T1d[i],self.time,Y[i])
			Yi.append(yitemp)
		return np.array(Yi)

# class pcasurrogate:
# 	def __init__(self,X,Y,norm=True,xbounds=None,xlabels=None,npca=50,pca_datafile=None):
# 		self.xnsamples,self.d = X.shape
# 		self.ynsamples,self.n = Y.shape
# 		self.xlabels = xlabels
# 		self.pca_datafile = pca_datafile
# 		assert self.xnsamples == self.ynsamples, "Sample mismatch between X and Y!"
# 		# normalize X data
# 		if norm == True:
# 			assert xbounds is not None, "Must provide bounds on X. Otherwise, set norm=False"
# 			LB,UB = xbounds
# 			self.Xscaled = 2*( (X - LB) * 1./(UB-LB) ) - 1
# 			# normalize Y -> [0,1]
# 			self.Ymax = np.amax(Y)
# 			self.Ymin = np.amin(Y)
# 			self.Yscaled = (Y - self.Ymin)/(self.Ymax - self.Ymin)
# 		# perform PCA with npca components
# 		self.npca = npca
# 		assert self.npca >= 2, "Number of PCA should be > 1!"
# 		self.project_flag = False
# 	def run(self,full=True,load=False,save=False,mre_cutoff=1e-3,K=None,error=False):
# 		# if load == True, just load data
# 		if error == False: assert K is not None, "Must choose K PCA modes."
# 		if K is not None: self.npca = K # set npca to K
# 		if save == True: assert self.pca_datafile is not None, "datafile field must be specified."
# 		if full == True and load == False:
# 			self.performPCA(self.npca)
# 			if error == True: # only perform recon error if user defined
# 				if self.pca_datafile is None or save == False:
# 					"running PCA without saving data..." 
# 					self.performPCAerroranalysis(save=False)
# 				else:
# 					self.performPCAerroranalysis(save=True,filename=self.pca_datafile)
# 		elif load == True:
# 			assert self.pca_datafile is not None, "Must specify pca_datafile."
# 			if save == True: warnings.warn("save option is ignored.")
# 			self.loaddata(self.pca_datafile)
# 		self.project(K=K,mre_cutoff=mre_cutoff)
# 	def performPCA(self,npca):
# 		print("performing PCA analysis...")
# 		self.pca = PCA(n_components=self.npca,svd_solver='full')
# 		self.pca.fit(self.Yscaled)
# 		# cumulative sum of explained variance
# 		self.explained_variance = np.cumsum(self.pca.explained_variance_ratio_)
# 		# save data
# 		self.muY = self.Yscaled.mean(0)
# 		self.Y0 = self.Yscaled - self.muY
# 		self.V = self.pca.components_.T # projection basis
# 		self.P = np.dot(self.Y0,self.V) # projection coefficients
# 		assert self.V.shape[0] == self.n, "Basis dimension should be same dimension as each Y sample."
# 		assert self.P.shape == (self.ynsamples,self.npca), "Shape of projection coefficients is not right!"
# 	def performPCAerroranalysis(self,save=True,filename=None):
# 		# calculate relative error of projections for increasing npca up to self.npca
# 		self.__mre = [] # mean relative reconstruction error
# 		for k in tqdm(range(2,self.npca)):
# 			Y0approx = np.dot(self.V[:,:k],self.P[:,:k].T).T
# 			RE = [] # relative error for each sample
# 			for i,y0 in enumerate(self.Y0):
# 				re = np.linalg.norm(y0 - Y0approx[i])/np.linalg.norm(self.muY + y0)
# 				RE.append(re)
# 			muRE = np.mean(np.array(RE))
# 			# print("Mean relative errors", np.mean(RE))
# 			self.__mre.append(muRE) # average over all samples
# 			self.mre = np.array(self.__mre)
# 		if save == True:
# 			assert filename is not None, "Must specify filename to save pca dictionary data."
# 			# save data as dictionary
# 			datadict = {}
# 			datadict['V'] = self.V # projection basis
# 			datadict['P'] = self.P # projection coefficients
# 			datadict['Y0'] = self.Y0 # centered data
# 			datadict['muY'] = self.muY # mean of original data
# 			datadict['mre'] = self.mre # reconstruction relative error
# 			datadict['expvar'] = self.explained_variance # cumulative explained variance
# 			datadict['Xscaled'] = self.Xscaled # normalized data to [-1,1]
# 			datadict['xlabels'] = self.xlabels # labels of x data
# 			print("saving PCA data...")
# 			with open(filename + '.pkl', 'wb') as pfile:
# 			    pickle.dump(datadict, pfile)
# 	def loaddata(self,filename):
# 		# import data dictionary with PCA modes and projections
# 		try:
# 			with open(filename, 'rb') as pfile:
# 				datadict = pickle.load(pfile)
# 		except:
# 			with open(filename + '.pkl', 'rb') as pfile:
# 				datadict = pickle.load(pfile)
# 		print("loading PCA data...")
# 		self.V = datadict['V'] # projection basis
# 		self.P = datadict['P'] # projection coefficients
# 		self.Y0 = datadict['Y0'] # centered data
# 		self.muY = datadict['muY'] # mean of original data
# 		self.mre = datadict['mre'] # reconstruction relative error
# 		self.explained_variance = datadict['expvar'] # cumulative explained variance
# 		self.Xscaled = datadict['Xscaled'] # normalized data to [-1,1]
# 		self.xlabels = datadict['xlabels'] # labels of x data
# 	def project(self,K=None,mre_cutoff=1e-3):
# 		if K is not None:
# 			self.K = K # use K if user defined
# 		else:
# 			# find PCA mode where recon error is greater than cutoff
# 			cutoff_indices = np.where(self.mre > mre_cutoff)[0]
# 			if len(cutoff_indices) == 0:
# 				warnings.warn("cutoff is too large. try decreasing. setting K = 1")
# 				self.K = 1 # only use 1 PCA mode
# 			else:
# 				self.K = cutoff_indices[-1] + 1 # add one since we are starting with 1 after the mean

# 		# get projections and basis columns
# 		self.Vk = self.V[:,:self.K]
# 		self.Pk = self.P[:,:self.K] # each row is the proj coef
# 		self.project_flag = True
# 	def invert(self,P):
# 		P = np.atleast_2d(P.copy())
# 		assert P.shape[1] == S.K, "dimension mismatch in P and V."
# 		R = np.dot(self.Vk,P.T)
# 		return R.T
# 	def pcefit(self):
# 		assert self.project_flag == True, "Must project data onto K PCA modes first. "
# 		npce = self.Vk.shape[1]


# # create function to plot sobol indices in a stacked graph
# # to be added to the pypce pceseries class
# C = []
# M = []
# Sobol = []
# t = []
# for pce in PCEs:
# 	C.append(pce['c'])
# 	M.append(pce['M'])
# 	t.append(pce['pca_mode'])
# 	Sobol.append(pce['S'])

# xlabels = np.array(['dens_sc', 'vel_sc', 'temp_sc', 
# 					'sigk1', 'sigk2', 'sigw1', 'sigw2',
# 					'beta_s', 'kap', 'a1', 'beta1r', 'beta2r'])
# pceS = pypce.pceseries(dim=xdim,C=C,MIs=M)

# Sobol = np.array(Sobol)
# # S = np.random.rand(4,12)
# normS = np.array([s/np.sum(s) for s in Sobol]).T

# import matplotlib.pyplot as plt
# import matplotlib._color_data as mcd

# xkcd_colors = []
# xkcd = {name for name in mcd.CSS4_COLORS if "xkcd:" + name in mcd.XKCD_COLORS}
# for j, n in enumerate(xkcd):
# 	xkcd = mcd.XKCD_COLORS["xkcd:" + n].upper()
# 	xkcd_colors.append(xkcd)

# Ps = []
# bottom = np.zeros(len(t))
# import seaborn as sns
# sns.set_palette(sns.color_palette("Paired", 12))
# plt.figure(figsize=(20,9))
# for ii in range(xdim):
# 	ptemp = plt.bar(t,normS[ii],bottom=bottom,width=.25)
# 	bottom = bottom + normS[ii] # reset bottom to new height
# 	Ps.append(ptemp)
# plt.ylabel('Sobol Scores')
# plt.ylim([0,1.1])
# # plt.title('Sobol indices by pca mode')
# plt.xticks(t, ('PCA1','PCA2','PCA3','PCA4','PCA5','PCA6'))
# plt.legend(((p[0] for p in Ps)), (l for l in xlabels),
# 			fancybox=True, shadow=True,
# 			loc='upper center', bbox_to_anchor=(0.5, 1.05),ncol=xdim)

# # plot explained variance
# plt.plot(t,S.explained_variance,'--ok')
# plt.savefig('figs/SA_'+y_label+'.png')


from .multiindex import RecursiveHypMultiIndex
from .multiindex import MultiIndex
from .pce import PCEBuilder
from .pce import PCEReg
from .quadrature import QuadGen
import tesuract.preprocessing
import tesuract.utils
# import tesuract.mpce
from .pce_multitarget_regression import RegressionWrapperCV, MRegressionWrapperCV, MPCEReg
import tesuract.experimental


import numpy as np
import matplotlib.pyplot as mpl
import pdb, ipdb, warnings, pickle
from numpy.polynomial.legendre import leggauss
from numpy.polynomial.hermite_e import hermegauss
import itertools
from scipy.special import comb

class QuadBase:
	def __init__(self,nquad):
		self.nquad = nquad
	def get1dQuad(self):
		pass

class LegendreQuad(QuadBase):
	def __init__(self,nquad=2):
		super().__init__(nquad)
	def get1dQuad(self,nquad=None):
		if nquad is not None: self.nquad = nquad
		x,w = leggauss(self.nquad)
		# rescale weights to sum to 1
		w = w/2.0
		return x,w

class HermiteQuad(QuadBase):
	''' normalized'''
	def __init__(self,nquad=2):
		super().__init__(nquad)
	def get1dQuad(self,nquad=None):
		if nquad is not None: self.nquad = nquad
		x,w = hermegauss(self.nquad)
		return x,w # add a factor of (2*np.pi)**-.5 to normalize each dimension

class ClenshawCurtis(QuadBase):
	def __init__(self,nquad=2):
		super().__init__(nquad)
	def _get1dQuad(self,nquad=None):
		''' old '''
		if nquad is not None: self.nquad = nquad
		if self.nquad == 1:
			return np.array([0.0]), np.array([2.0])
		else:
			n = self.nquad
			x = np.cos(np.pi*(n-1-np.arange(n))/(n-1))
			w = np.ones(len(x))
			for i in range(n):
				theta = i*np.pi/(n-1)
				for j in range(1,int(.5*(n-1)+1)):
					if 2*j == n-1:
						f = 1.0
					else:
						f = 2.0
					# print(i,j,f)
					w[i] -= f*np.cos(2.0*j*theta)/(4.0*j**2-1)
			w[0] /= n-1
			w[1:-1] = 2*w[1:-1]/(n-1)
			w[-1] *= 1.0/(n-1)
			return x,w
	def get1dQuad(self,nquad=None):
		'''from chaospy'''
		if nquad is not None: 
			self.nquad = nquad
		degree = self.nquad
		n = self.nquad
		if n == 1:
			points = np.array([0.0])
			weights = np.array([2.0])
		else:
			points = -np.cos((np.pi * np.arange(n)) / (n - 1))
			if n == 2:
				weights = np.array([1.0, 1.0])
			else:
				n -= 1
				N = np.arange(1, n, 2)
				length = len(N)
				m = n - length
				v0 = np.concatenate(
					[2.0 / N / (N - 2), np.array([1.0 / N[-1]]), np.zeros(m)]
				)
				v2 = -v0[:-1] - v0[:0:-1]
				g0 = -np.ones(n)
				g0[length] += n
				g0[m] += n
				g = g0 / (n ** 2 - 1 + (n % 2))

				w = np.fft.ihfft(v2 + g)
				assert max(w.imag) < 1.0e-15
				w = w.real

				if n % 2 == 1:
					weights = np.concatenate([w, w[::-1]])
				else:
					weights = np.concatenate([w, w[len(w) - 2 :: -1]])
		weights = weights/2.0
		return points, weights

class QuadFactory:
	# generates QuadBase class object
	@staticmethod
	def newQuad(quadtype='legendre_gauss'):
		if quadtype == 'legendre_gauss':
			Q = LegendreQuad()
		if quadtype == 'clenshaw_curtis':
			Q = ClenshawCurtis()
		if quadtype == 'hermite_gauss':
			Q = HermiteQuad()
		return Q

class QuadRule:
	def __init__(self,x,w):
		self.x = x
		self.w = w
		self.n = len(w)
		if x.ndim == 1:
			self.dim = 1
			self.x = np.atleast_2d(x).T # col vector
		else:
			self.dim = x.shape[1]
		assert len(x) == len(w), "x and w dont habe the same # of points"
	def __add__(self,other):
		assert self.dim == other.dim, "Dimensions do not match!"
		xnew = np.vstack([self.x,other.x])
		wnew = np.hstack([self.w,other.w])
		Qnew = QuadRule(xnew,wnew)
		return Qnew
	def __sub__(self,other):
		assert self.dim == other.dim, "Dimensions do not match!"
		xnew = np.vstack([self.x,other.x])
		wnew = np.hstack([self.w,-1*other.w])
		Qnew = QuadRule(xnew,wnew)
		return Qnew
	def __mul__(self,other):
		# tensor product
		index_comb = list(itertools.product(range(self.n),range(other.n)))
		xnew = [np.concatenate([self.x[i[0]],other.x[i[1]]]) for i in index_comb]
		wnew = [self.w[i[0]]*other.w[i[1]] for i in index_comb]
		Qnew = QuadRule(np.array(xnew),np.array(wnew))
		return Qnew
	def copy(self):
		return QuadRule(self.x,self.w)

class QuadOps:
	@staticmethod
	def getMultiIndexLevel(level,ndim):
		''' returns the multindices of order = level'''
		iup = 0
		nup_level = int(comb(ndim+level-1,level))
		M = np.zeros((nup_level,ndim))
		if ndim == 1:
			M[0,0] = level
		else:
			for first in range(level,-1,-1):
				theRest = QuadOps.getMultiIndexLevel(level-first,ndim-1)

				for j in range(len(theRest)):
					# print(iup,j)
					M[iup,0] = first
					M[iup,1:ndim] = theRest[j,0:ndim-1]
					iup += 1

		return M
	@staticmethod
	def compressRule(Q):
		# assert self.rule_ is not None, "Must set rule first."
		# convert numpy array to list of tuples
		xtuple = [tuple(xi) for xi in Q.x]
		# create a dictionary 
		from collections import defaultdict
		dd = defaultdict(list)
		for ii,xi in enumerate(xtuple):
			dd[xi].append(Q.w[ii])
		# sum weights over keys
		for key in dd:
			dd[key] = np.sum(dd[key])
		x = np.array(list(dd.keys()))
		w = np.array([dd[key] for key in dd])
		x = x[np.abs(w) > 1e-12]
		w = w[np.abs(w) > 1e-12]
		return QuadRule(x,w)

# deprecated for sparse, but good for full tensor product grid
class QuadBuilder():
	def __init__(self,grid_type='sparse',order=2,quad_type='Legendre'):
		self.grid_type = grid_type
		self.quad_type = quad_type
		self.order = order
		self.ndim = None
		self.growth_rule = None
	def SetRule(self,ndim):
		self.ndim = ndim
		if self.grid_type == 'full':
			self._full()
		if self.grid_type == 'sparse':
			if self.quad_type == 'legendre_gauss': 
				self.growth_rule = 0
			self._sparse()
		return self
	def _full(self):
		# cannot do mixed quad yet. Easy if quad type takes in array
		quad_gen = QuadFactory.newQuad(self.quad_type)
		x,w = quad_gen.get1dQuad(nquad=self.order+1) # 0th order means 1 point
		q1d = QuadRule(x,w)
		qnew = q1d.copy()
		for i in range(1,self.ndim):
			qnew = qnew*q1d
		q = qnew.copy()
		self.rule_ = q
	def _sparse(self):
		for nlevel in range(-1,self.order):
			self._SetNextLevel2(nlevel)
	def _SetNextLevel(self,nlevel):
		nlevel += 1
		M = QuadOps.getMultiIndexLevel(nlevel,self.ndim)
		nM = M.shape[0]
		M_npts = np.zeros((nM,self.ndim))
		quad_gen = QuadFactory.newQuad(self.quad_type)
		for j in range(nM):
			Mj = M[j] # jth row of the multiindexlevel
			# 1 if Mj == 0, 3 if 1, else (Mj_i+1)^2
			if self.growth_rule == 0:
				npts = 1*(Mj == 0) + 3*(Mj == 1) + ((Mj)**2+1)*(Mj > 1)
				npts_1 = 0*(Mj == 0) + 1*(Mj == 1) + ((Mj-1)**2+1)*(Mj > 1)
			elif self.growth_rule == 1:
				npts = ((Mj+1)**2-1)*(Mj > 1)
				npts_1 = (Mj**2-1)*(Mj > 1)
			npts = npts.astype('int')
			npts_1 = npts_1.astype('int')
			print(npts,npts_1)
			xw = [quad_gen.get1dQuad(nquad=int(n)) for n in list(npts)]
			rules = [QuadRule(xwi[0],xwi[1]) for xwi in xw]
			xw_1 = [quad_gen.get1dQuad(nquad=int(n_1)) for n_1 in list(npts_1)]
			rules_1 = [QuadRule(xwi_1[0],xwi_1[1]) for xwi_1 in xw_1]
			srules = []
			for ii in range(len(npts)):
				if npts_1[ii] > 0:
					srules.append(rules[ii] - rules_1[ii])
				else:
					srules.append(rules[ii])

			# multiply rules in srules
			r = srules[0].copy()
			for ri in srules[1:]:
				r = r*ri

			if j == 0:
				rule_level = r.copy()
			else:
				rule_level = r + rule_cur
			rule_cur = rule_level.copy()
		# pdb.set_trace()
		if nlevel == 0:
			rule_total = rule_level.copy()
		else:
			rule_total = self.rule_ + rule_level
		self.rule_ = rule_total.copy()
		return self
	def _SetNextLevel2(self,nlevel):
		nlevel += 1
		M = QuadOps.getMultiIndexLevel(nlevel,self.ndim)
		self.M = M
		# nM = M.shape[0]
		M_npts = np.zeros(M.shape)
		quad_gen = QuadFactory.newQuad(self.quad_type)
		for j in range(len(M)):
			rules = []
			rules_1 = []
			srules = []
			Mj = M[j]
			for id in range(self.ndim):
				if M[j,id] == 0:
					npts = 1
					npts_1 = 0
				elif M[j,id] == 1:
					npts = 3
					npts_1 = 1
				else:
					npts = int((M[j,id])**2) + 1
					npts_1 = int((M[j,id]-1)**2)+1

				x,w = quad_gen.get1dQuad(nquad=npts)
				rule = QuadRule(x,w)
				if npts_1 > 0:
					x_1,w_1 = quad_gen.get1dQuad(nquad=npts_1)
					rule_1 = QuadRule(x_1,w_1)
					srule = rule - rule_1
				else:
					srule = rule.copy()
				srules.append(srule)
				# end of id iterator
			# multiple rule
			rule_temp = srules[0].copy()
			for s in srules[1:]:
				rule_temp = rule_temp * s
			# rule_temp = srules[0]*srules[1]
			if j == 0:
				rule_level = rule_temp.copy()
			else:
				rule_level = rule_cur + rule_temp
			# pdb.set_trace()
			rule_cur = rule_level.copy()

			# end of j iterator

		if nlevel == 0:
			rule_total = rule_level.copy()
		else:
			rule_total = self.rule_ + rule_level
		self.rule_ = rule_total.copy()
		# test = np.unique(self.rule_.x,axis=0)
		# ind = []
		# for i in range(len(test)):
		# 	ind = np.array_equal(test[0],self.rule_.x[i])
		# 	if ind == True:
		# 		print(self.rule_.w[i])
		# ipdb.set_trace()
		self.rule_ = QuadOps.compressRule(self.rule_)
		return self

def construct_lookup(orders, dim, rules='gaussian'):
	"""
	Create abscissas and weights look-up table so values do not need to be
	re-calculatated on the fly.
	"""
	if isinstance(rules, str):
		rules = (rules,)*dim
	if isinstance(orders, int):
		orders = orders*np.ones(dim, dtype=int)
	x_lookup = []
	w_lookup = []
	if rules[0] == "gaussian" or "legendre_gauss":
		Q = QuadFactory.newQuad("legendre_gauss")
		growth = False
	if rules[0] == "clenshaw_curtis":
		Q = QuadFactory.newQuad("clenshaw_curtis")
		growth = True
	# if rules[0] == 'hermite':
		# Q = QuadFactory.newQuad("hermite_gauss")
	for max_order, rule in zip(orders, rules):
		x_lookup.append([])
		w_lookup.append([])
		for order in range(max_order+1):
			if growth == True:
				order_adj = 2**(order)+1
			else:
				order_adj = order + 1
			abscissas, weights = Q.get1dQuad(order_adj)
			x_lookup[-1].append(abscissas)
			w_lookup[-1].append(weights)
	return x_lookup, w_lookup


def construct_collection(orders, dim, x_lookup, w_lookup
                         ):
	"""Create a collection of {abscissa: weight} key-value pairs."""
	if isinstance(orders, int):
		orders = orders*np.ones(dim, dtype=int)
	order = np.min(orders)
	skew = orders-order

	# Indices and coefficients used in the calculations
	# indices = chaospy.numpoly.glexindex(
	# 	order-len(dists)+1, order+1, dimensions=len(dists))
	mi = []
	for ilevel in range(order-dim+1, order+1):
		mi.append(QuadOps.getMultiIndexLevel(ilevel, dim))
	indices = np.vstack(mi).astype('int')
	coeffs = np.sum(indices, -1)
	coeffs = (2*((order-coeffs+1) % 2)-1)*comb(dim-1, order-coeffs)

	collection = defaultdict(float)
	for bidx, coeff in zip(indices+skew, coeffs.tolist()):
		abscissas = [value[idx] for idx, value in zip(bidx, x_lookup)]
		weights = [value[idx] for idx, value in zip(bidx, w_lookup)]
		for abscissa, weight in zip(product(*abscissas), product(*weights)):
			collection[abscissa] += np.prod(weight)*coeff

	return collection

from collections import defaultdict
from itertools import product

class QuadGen(object):
	def __init__(self,dim,order=2,rule='legendre_gauss',sparse=True):
		self.dim = dim
		self.order = order
		self.rule = rule
		self.sparse = sparse
	def run(self,order=None):
		if order is None: order = self.order
		if self.sparse == False:
			Qtemp = QuadBuilder(order=order,grid_type='full',quad_type=self.rule)
			Qtemp.SetRule(ndim=self.dim)
			x,w = Qtemp.rule_.x,Qtemp.rule_.w
		elif self.sparse == True:
			# create initial lookup table of 1d quadratures
			self.x_lookup, self.w_lookup = construct_lookup(orders=order,dim=self.dim,rules=self.rule)
			self.collection = construct_collection(orders=order,dim=self.dim,
								x_lookup=self.x_lookup,w_lookup=self.w_lookup)
			# scale and renormalize to over [0,1] and unit weight
			x = sorted(self.collection)
			w = np.array([self.collection[key] for key in x])
			x = np.array(x) # each row is a data point
			assert (np.abs(np.sum(w) - 1.0)) <= 1e-12, "Weights are not normalized on [0,1]." 
		x = .5*(x + 1)
		assert np.all(x >= 0.0) and np.all(x <= 1.0), "Points are outside the range [0,1]^d"
		self.points, self.weights = x,w
		return self.points, self.weights


# test...
dim = 5
Q = QuadGen(dim=dim,order=5,rule='legendre_gauss',sparse=True)
x,w = Q.run()

Q2 = QuadGen(dim=dim,order=5,rule='legendre_gauss',sparse=False)
x2,w2 = Q2.run()

# genz function tests

def genz_exp(x):
	if x.ndim == 2:
		output = np.exp(np.sum(x - 1,axis=1))
	elif x.ndim == 1:
		output = np.exp(np.sum(x - 1))
	return output
soln = (np.exp(0) - np.exp(-1.0))**dim

approx = np.sum(genz_exp(x)*w)
error = np.mean((soln - approx)**2)/np.mean(soln**2)
print(soln, 'vs', approx,' , error = ', error)












# See https://michaelgoerz.net/notes/extending-sphinx-napoleon-docstring-sections.html

# # -- Fixing bug with google docs showing attributes-------------
from sphinx.ext.napoleon.docstring import GoogleDocstring

# first, we define new methods for any new sections and add them to the class
def parse_keys_section(self, section):
    return self._format_fields('Keys', self._consume_fields())
GoogleDocstring._parse_keys_section = parse_keys_section

def parse_attributes_section(self, section):
    return self._format_fields('Attributes', self._consume_fields())
GoogleDocstring._parse_attributes_section = parse_attributes_section

def parse_class_attributes_section(self, section):
    return self._format_fields('Class Attributes', self._consume_fields())
GoogleDocstring._parse_class_attributes_section = parse_class_attributes_section

# we now patch the parse method to guarantee that the the above methods are
# assigned to the _section dict
def patched_parse(self):
    self._sections['keys'] = self._parse_keys_section
    self._sections['class attributes'] = self._parse_class_attributes_section
    self._unpatched_parse()
GoogleDocstring._unpatched_parse = GoogleDocstring._parse
GoogleDocstring._parse = patched_parse


from setuptools import setup


def readme():
    with open("README.rst") as f:
        return f.read()


setup(
    name="tesuract",
    version="0.2.0",
    description="Tensor surrogate construction methods",
    long_description="UQ and ML tools for surrogate construction",
    url="https://github.com/kennychowdhary/tesuract",
    author="Kenny Chowdhary",
    author_email="kchowdh@sandia.gov",
    license="BSD3",
    packages=["tesuract"],
    test_suite="nose.collector",
    tests_required=["nose"],
    install_requires=["numpy", "sklearn", "scipy", "tqdm"],
    include_package_data=True,
    zip_safe=False,
)



import numpy as np 

from pce import PCEReg, PCEBuilder
from sklearn.datasets import make_regression
from sklearn import preprocessing


rn = np.random.RandomState(123)
X,y = make_regression(n_samples=2000,n_features=12,random_state=rn)

feature_scaler = preprocessing.MinMaxScaler(feature_range=(-1+1e-6,1-1e-6))
X = feature_scaler.fit_transform(X)

p = PCEReg(order=4,fit_type='LassoCV')
p.fit(X,y)

